#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2014  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
"""
Paperwork configuration management code
"""

import collections
import logging

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Setting(object):
    def __init__(self, core, section, token, default_value_func):
        self.core = core
        self.section = section
        self.token = token
        self.default_value_func = default_value_func

    def get(self):
        value = self.core.call_success(
            "config_backend_get", self.section, self.token, None
        )
        if value is None:
            return self.default_value_func()
        else:
            return value

    def put(self, value):
        self.core.call_all(
            "config_backend_put", self.section, self.token, value
        )


class Plugin(PluginBase):
    """
    Translate values from the configuration into more usable ones.
    Provides default values (except for plugins).
    """

    def __init__(self):
        self.core = None
        self.settings = {}
        self.values = {}
        # applicatiom here is a bit more specific: paperwork-gtk,
        # paperwork-shell, etc.
        # It is used to known which plugin list must be loaded
        self.plugin_list_name = None
        self.observers = collections.defaultdict(set)

    def get_interfaces(self):
        return ['config']

    def get_deps(self):
        return [
            {
                'interface': 'app',
                'defaults': ['openpaperwork_core.app'],
            },
            {
                'interface': 'config_backend',
                'defaults': ['openpaperwork_core.config.backend.configparser'],
            },
        ]

    def init(self, core):
        self.core = core
        self.settings = {}

    def config_load(self):
        application = self.core.call_success("app_get_fs_name")
        LOGGER.info("Loading configuration for %s", application)

        self.values = {}
        self.core.call_all('config_backend_load', application)
        for observers in self.observers.values():
            for observer in observers:
                observer()

    def config_load_plugins(self, plugin_list_name, default_plugins=[]):
        LOGGER.info("Loading plugins for '%s'", plugin_list_name)
        self.plugin_list_name = plugin_list_name
        self.core.call_all(
            'config_backend_load_plugins', plugin_list_name, default_plugins
        )

    def config_get_plugin_list_name(self):
        return self.plugin_list_name

    def config_save(self):
        LOGGER.info("Saving configuration")
        self.core.call_all('config_backend_save')

    def config_build_simple(
                self, section, token, default_value_func
            ):
        """
        Provide a default simple implementation for a new setting that can
        be registered using 'config_register'.

        Arguments:
          - section: Section in which option must be stored (see ConfigParser)
          - token: token name for the option (see ConfigParser)
          - default_value_func: function to call to get the default value if
            none is stored in the file.
        """
        return Setting(self.core, section, token, default_value_func)

    def config_register(self, key: str, setting):
        """
        Add another setting to manage. Make this setting available to other
        components.

        Arguments:
          - key: configuration key
          - setting: Setting must be an object providing a method `get()`
            and a method `put(value)`. See :func:`config_build_simple` to
            get quickly a default implementation.
        """
        LOGGER.debug("Registering configuration: %s", key)
        self.settings[key] = setting

    def config_list_options(self):
        return list(self.settings.keys())

    def config_get_setting(self, key: str):
        return self.settings[key]

    def config_get(self, key: str):
        LOGGER.debug("Config get: %s", key)
        if key not in self.settings:
            return None
        if key not in self.values:
            self.values[key] = self.settings[key].get()
        return self.values[key]

    def config_get_default(self, key: str):
        return self.settings[key].default_value_func()

    def config_put(self, key: str, value):
        """
        Store a setting value.
        Warning: You must call :func:`config_save` so the changes are actually
        saved.

        Arguments:
          - key: configuration key,
          - value: can be of many types (`str`, `int`, etc).
        """
        LOGGER.debug("Config put: %s", key)
        self.settings[key].put(value)
        self.values[key] = value
        if key in self.observers:
            for callback in self.observers[key]:
                callback()

    def config_add_plugin(self, plugin, plugin_list_name=None):
        if plugin_list_name is None:
            plugin_list_name = self.plugin_list_name
        LOGGER.debug("Config add plugin: %s -> %s", plugin, plugin_list_name)
        self.core.call_all(
            'config_backend_add_plugin', plugin_list_name, plugin
        )

    def config_remove_plugin(self, plugin, plugin_list_name=None):
        if plugin_list_name is None:
            plugin_list_name = self.plugin_list_name
        LOGGER.debug(
            "Config remove plugin: %s -> %s",
            plugin, plugin_list_name
        )
        self.core.call_all(
            'config_backend_remove_plugin', plugin_list_name, plugin
        )

    def config_list_plugins(self, plugin_list_name=None):
        if plugin_list_name is None:
            plugin_list_name = self.plugin_list_name
        return self.core.call_success(
            "config_backend_list_active_plugins", plugin_list_name
        )

    def config_reset_plugins(self, plugin_list_name=None):
        if plugin_list_name is None:
            plugin_list_name = self.plugin_list_name
        LOGGER.debug("Config reset plugin: %s", plugin_list_name)
        return self.core.call_success(
            "config_backend_reset_plugins", plugin_list_name
        )

    def config_add_observer(self, key: str, callback):
        self.observers[key].add(callback)

    def config_remove_observer(self, key: str, callback):
        self.observers[key].remove(callback)



"""
Select a scanner if none has been selected yet or if the currently selected one
doesn't exist anymore.

IMPORTANT: This plugin must list the devices only if required. Otherwise
it tends to mess up some Sane backend like the Brother one for instance.
"""
import logging

import openpaperwork_core

import paperwork_backend.util


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def get_interfaces(self):
        return []

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
        ]

    def init(self, core):
        super().init(core)

    def scan_get_scanner_promise(self, scanner_dev_id=None, autoselect=True):
        """
        Overload the method 'scan_get_scanner_promise' to check whether
        the device actually exists, and if not find the closest one.
        """
        if not autoselect:
            return None

        def get_scanner(scanner_dev_id=None):
            scanner = None
            if scanner_dev_id is None:
                scanner_dev_id = self.core.call_success(
                    "config_get", "scanner_dev_id"
                )
            try:
                if scanner_dev_id is not None:
                    scanner = self.core.call_success(
                        "scan_get_scanner", scanner_dev_id
                    )
            except Exception as exc:
                # may happen if ID has changed (those containing the USB
                # bus + ID for instance)
                LOGGER.warning(
                    "Failed to get scanner '%s'."
                    " Will try to find the closest one",
                    scanner_dev_id, exc_info=exc
                )
            if scanner is not None:
                return scanner

            devs = self.core.call_success("scan_list_scanners")
            scanner_dev_id = self._find_dev_id(devs, scanner_dev_id)
            if scanner_dev_id is None:
                return None
            return self.core.call_success("scan_get_scanner", scanner_dev_id)

        return openpaperwork_core.promise.ThreadedPromise(
            self.core, get_scanner, args=(scanner_dev_id,)
        )

    def _find_dev_id(self, devs, current_dev_id=None):
        devs = [dev[0] for dev in devs]
        LOGGER.info("Available scanners: %s", devs)

        if len(devs) <= 0:
            LOGGER.error("No scanner found !")
            return None
        elif current_dev_id is None:
            # pick a scanner at random.
            selected = devs[0]
        else:
            LOGGER.info("Previously selected scanner: %s", current_dev_id)
            # look for the closest scanner ID
            devs = [
                (
                    paperwork_backend.util.levenshtein_distance(
                        dev, current_dev_id
                    ),
                    dev
                )
                for dev in devs
            ]
            devs.sort()
            selected = devs[0][1]

        LOGGER.info("Selected scanner: %s", selected)
        # ASSUMPTION(Jflesch): we assume here the requested scanner was
        # the one selected in the configuration --> we update the configuration
        # to avoid listing devices every time.
        self.core.call_success("config_put", "scanner_dev_id", selected)
        return selected



import unittest

from .. import (Core, promise)


class AbstractTestCallback(unittest.TestCase):
    def get_plugin_name(self):
        """
        Must be overloaded by subclasses
        """
        assert False

    def setUp(self):
        self.core = Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

        self.val = None

    def test_basic(self):
        def set_val(value):
            self.val = value

        # queue some calls
        self.core.call_all("mainloop_schedule", set_val, 22)
        self.core.call_all("mainloop_quit_graceful")

        self.core.call_one('mainloop')

        self.assertEqual(self.val, 22)


class AbstractTestPromise(unittest.TestCase):
    def get_plugin_name(self):
        """
        Must be overloaded by subclasses
        """
        assert False

    def setUp(self):
        self.core = Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

        self.alpha_called = False
        self.beta_called = False
        self.stop_called = False
        self.exc_raised = False
        self.idx = 0

    def test_single(self):
        self.stop_called = False

        def stop():
            self.stop_called = True

        p = promise.Promise(self.core, stop)
        p.schedule()
        self.core.call_all("mainloop_quit_graceful")

        self.core.call_one("mainloop")
        self.assertTrue(self.stop_called)

    def test_chain(self):
        self.alpha_called = -1
        self.beta_called = -1
        self.stop_called = -1
        self.idx = 0

        def alpha():
            self.alpha_called = self.idx
            self.idx += 1
            return "alpha"

        def beta(previous):
            self.assertEqual(previous, "alpha")
            self.beta_called = self.idx
            self.idx += 1

        def stop():
            self.stop_called = self.idx
            self.idx += 1

        p = promise.Promise(self.core, alpha)
        p = p.then(beta)
        p = p.then(stop)
        p.schedule()
        self.core.call_all("mainloop_quit_graceful")

        self.core.call_one("mainloop")
        self.assertEqual(self.alpha_called, 0)
        self.assertEqual(self.beta_called, 1)
        self.assertEqual(self.stop_called, 2)

    def test_catch(self):
        self.alpha_called = False
        self.beta_called = False
        self.stop_called = False
        self.exc_raised = False

        def alpha():
            self.alpha_called = True

        def beta():
            self.beta_called = True
            raise Exception("paf")

        def stop():
            self.stop_called = True

        def on_exc(exc):
            self.exc_raised = True

        p = promise.Promise(self.core, alpha)
        p = p.then(beta)
        p = p.then(stop)
        p = p.catch(on_exc)
        p.hide_caught_exceptions = True
        p.schedule()
        self.core.call_all("mainloop_quit_graceful")

        self.core.call_one("mainloop")
        self.assertTrue(self.alpha_called)
        self.assertTrue(self.beta_called)
        self.assertFalse(self.stop_called)
        self.assertTrue(self.exc_raised)



import logging
import sys
import threading
import time

import openpaperwork_core
import openpaperwork_core.deps


# TODO(Jflesch): refactor with openpaperwork_gtk.widgets.progress


LOGGER = logging.getLogger(__name__)

TIME_BETWEEN_UPDATES = 0.3
# Tasks are often chained one after the other. We don't want the button/popover
# to disappear and reappear continually. So when a task ends, we
# give them some extra time to live.
# STAY_ALIVES is the number of updates we wait before hiding them.
STAY_ALIVES = int(2.0 / TIME_BETWEEN_UPDATES)


class ProgressDialog:
    def __init__(self, plugin):
        LOGGER.info("Opening dialog 'wait for background tasks'")

        self.plugin = plugin
        self.core = plugin.core

        self.progress_widget_trees = {}

        self.dialog_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.wait_for_background_tasks",
            "wait_for_background_tasks.glade"
        )
        if self.dialog_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.dialog_tree.get_object("dialog").connect(
            "response", self._suicide
        )
        self.dialog_tree.get_object("dialog").show_all()

        # A thread updates the widgets every 300ms. We don't update them
        # each time on_progress() is called to not degrade performances
        self.thread = threading.Thread(target=self._thread)
        self.thread.daemon = True
        self.thread.start()

    def _thread(self):
        self.stay_alives = STAY_ALIVES
        while self.thread is not None:
            time.sleep(TIME_BETWEEN_UPDATES)
            self.core.call_one(
                "mainloop_execute", self._upd_progress_widgets
            )

    def _upd_progress_widgets(self):
        with self.plugin.lock:
            progresses = self.plugin.progresses.items()
            self.plugin.progresses = {}
            for (upd_type, (progress, description)) in progresses:
                self._upd_progress_widget(
                    upd_type, progress, description
                )

            if len(self.progress_widget_trees) > 0:
                self.stay_alives = STAY_ALIVES
                return

            if self.stay_alives > 0:
                self.stay_alives -= 1
                return

            self.thread = None
            self.dialog_tree.get_object("dialog").destroy()

    def _upd_progress_widget(self, upd_type, progress, description):
        box = self.dialog_tree.get_object("task_list")

        if progress >= 1.0:  # deletion of progress
            if upd_type not in self.progress_widget_trees:
                return

            LOGGER.info("Task '%s' has ended", upd_type)
            widget_tree = self.progress_widget_trees.pop(upd_type)
            details = widget_tree.get_object("progress_bar")
            box.remove(details)
            details.unparent()

            LOGGER.info(
                "Task '%s' has ended (%d remaining)",
                upd_type, len(self.progress_widget_trees)
            )
            return

        if upd_type not in self.progress_widget_trees:
            # creation of progress widget
            LOGGER.info(
                "Task '%s' has started (%d already active)",
                upd_type, len(self.progress_widget_trees)
            )
            # reuse the progress widgets from
            # openpaperwork_gtk.widgets.progress
            widget_tree = self.core.call_success(
                "gtk_load_widget_tree",
                "openpaperwork_gtk.widgets.progress",
                "progress_details.glade"
            )
            box.add(widget_tree.get_object("progress_bar"))
            self.progress_widget_trees[upd_type] = widget_tree

        else:
            widget_tree = self.progress_widget_trees[upd_type]

        # update of progress
        progress_bar = widget_tree.get_object("progress_bar")
        progress_bar.set_fraction(progress)
        progress_bar.set_text(description if description is not None else "")

    def _suicide(self, *args, **kwargs):
        LOGGER.warning("User requested a forced quit")
        # In theory, the following shouldn't hurt any more than a power outtage
        # and is really reliable way to end Paperwork
        sys.exit(0)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.dialog = None
        self.progresses = {}
        self.lock = threading.Lock()

    def get_interfaces(self):
        return [
            'progress_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
        ]

    def _mainloop_quit_graceful(self):
        self.dialog = ProgressDialog(self)

    def mainloop_quit_graceful(self):
        self.core.call_one("mainloop_execute", self._mainloop_quit_graceful)

    def on_progress(self, upd_type, progress, description=None):
        with self.lock:
            if progress > 1.0:
                LOGGER.warning(
                    "Invalid progression (%f) for [%s]",
                    progress, upd_type
                )
                progress = 1.0

            self.progresses[upd_type] = (progress, description)



import datetime
import logging
import os

import openpaperwork_core


LOGGER = logging.getLogger(__name__)

DOCNAME_FORMAT = "%Y%m%d_%H%M_%S"


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -1000  # see page_delete_by_url() / page_move_by_url()

    def get_interfaces(self):
        return [
            "document_storage",
            "stats"
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
        ]

    def init(self, core):
        super().init(core)
        setting = self.core.call_success(
            "config_build_simple", "Global", "WorkDirectory",
            lambda: self.core.call_success(
                "fs_safe", os.path.expanduser("~/papers")
            )
        )
        self.core.call_all("config_register", "workdir", setting)
        self.core.call_all(
            "config_add_observer", "workdir", self._on_work_dir_changed
        )

    def _on_work_dir_changed(self):
        LOGGER.info("Work directory has changed")
        self.core.call_all("on_storage_changed")

    def storage_get_id(self):
        """
        Returns a string identifying the storage (work directory)
        currently used.

        Do not assume it's a valid directory path. For instance, this plugin
        could be replaced by a MariaDB database someday, and so the string
        would identify the database instead of a directory.
        """
        return self.core.call_success('config_get', 'workdir')

    def storage_get_all_docs(self, out: list, only_valid=True):
        """
        Returns all document IDs and URLs in the work directory
        """
        workdir = self.core.call_success('config_get', 'workdir')
        if self.core.call_success('fs_exists', workdir) is None:
            # we are not the plugin handling this storage (?)
            return
        LOGGER.info("Loading document list from %s", workdir)
        nb = 0
        for doc_url in self.core.call_success('fs_listdir', workdir):
            doc_id = self.core.call_success("fs_basename", doc_url)
            if doc_id.startswith("."):
                continue
            if only_valid and not self.core.call_success("fs_isdir", doc_url):
                LOGGER.warning(
                    "Ignoring document '%s': not a directory", doc_url
                )
                continue
            if (
                    only_valid and
                    self.core.call_success("is_doc", doc_url) is None):
                LOGGER.warning(
                    "Ignoring document '%s': not recognized document type",
                    doc_url
                )
                continue
            out.append((doc_id, doc_url))
            nb += 1
        LOGGER.info("%d documents found in %s", nb, workdir)

    def doc_id_to_url(self, doc_id, existing=True):
        assert doc_id is not None
        workdir = self.core.call_success('config_get', 'workdir')
        url = self.core.call_success("fs_join", workdir, doc_id)
        if existing and self.core.call_success("fs_isdir", url) is None:
            return None
        return url

    def doc_url_to_id(self, doc_url):
        workdir = self.core.call_success('config_get', 'workdir')
        if not doc_url.startswith(workdir):
            return None
        return self.core.call_success("fs_basename", doc_url)

    def doc_get_date_by_id(self, doc_id):
        # Doc id is expected to have this format:
        # YYYYMMDD_hhmm_ss_NN_something_else
        doc_id = doc_id.split("_", 3)
        doc_id = "_".join(doc_id[:3])
        try:
            return datetime.datetime.strptime(doc_id, DOCNAME_FORMAT)
        except ValueError:
            return None

    def doc_get_id_by_date(self, date):
        return date.strftime(DOCNAME_FORMAT)

    # datetime.datetime.now cannot be mocked with unittest.mock.patch
    # (datetime is built-in) --> allow dependency injection here
    def storage_get_new_doc(self, now_func=datetime.datetime.now):
        workdir = self.core.call_success('config_get', 'workdir')

        base_doc_id = now_func().strftime(DOCNAME_FORMAT)
        base_doc_url = self.core.call_success("fs_join", workdir, base_doc_id)

        doc_id = base_doc_id
        doc_url = base_doc_url
        doc_idx = 0

        while self.core.call_success("fs_exists", doc_url) is not None:
            doc_idx += 1
            doc_id = "{}_{}".format(base_doc_id, doc_idx)
            doc_url = "{}_{}".format(base_doc_url, doc_idx)

        return (doc_id, doc_url)

    def storage_delete_doc_id(self, doc_id, trash=True):
        doc_url = self.doc_id_to_url(doc_id)
        if doc_url is None:
            return
        self.core.call_success("fs_rm_rf", doc_url, trash=trash)

    def stats_get(self, stats):
        LOGGER.info("Counting documents for statistics...")
        all_docs = []
        self.storage_get_all_docs(all_docs)
        stats['nb_documents'] += len(all_docs)

    def page_delete_by_url(self, doc_url, page_idx):
        workdir = self.core.call_success('config_get', 'workdir')
        if not doc_url.startswith(workdir):
            return None
        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is not None and nb_pages > 0:
            return
        LOGGER.warning(
            "All pages of document %s have been removed. Removing document",
            doc_url
        )
        self.core.call_success("fs_rm_rf", doc_url)

    def page_move_by_url(
                self,
                source_doc_url, source_page_idx,
                dest_doc_url, dest_page_idx
            ):
        workdir = self.core.call_success('config_get', 'workdir')
        if not source_doc_url.startswith(workdir):
            return None
        workdir = self.core.call_success('config_get', 'workdir')
        if not dest_doc_url.startswith(workdir):
            return None

        nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", source_doc_url
        )
        if nb_pages is not None and nb_pages > 0:
            return
        LOGGER.warning(
            "All pages of document %s have been removed. Removing document",
            source_doc_url
        )
        self.core.call_success("fs_rm_rf", source_doc_url)

    def doc_rename_by_url(self, src_doc_url, dst_doc_url):
        if not self.core.call_success("fs_isdir", src_doc_url):
            # May happen on integrated documentation PDF files
            LOGGER.warning("Cannot rename non-directory documents")
            return
        idx = 0
        dst_url = dst_doc_url
        while self.core.call_success("fs_exists", dst_url) is not None:
            idx += 1
            dst_url = "{}_{}".format(dst_doc_url, idx)
        self.core.call_success("fs_rename", src_doc_url, dst_url)
        return dst_url



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import openpaperwork_core

from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "document_storage",
                "defaults": ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                "interface": "pages",
                "defaults": [
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.thumbnail',
                ],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'move', help=_("Move a page")
        )
        p.add_argument(
            'source_doc_id',
            help=_("Source document")
        )
        p.add_argument(
            'source_page', type=int,
            help=_("Page to move")
        )
        p.add_argument(
            'dest_doc_id',
            help=_("Destination document")
        )
        p.add_argument(
            'dest_page', type=int,
            help=_("Target page number")
        )

    def cmd_run(self, console, args):
        if args.command != 'move':
            return None

        source_doc_id = args.source_doc_id
        source_page_idx = args.source_page - 1
        dest_doc_id = args.dest_doc_id
        dest_page_idx = args.dest_page - 1

        source_doc_url = self.core.call_success("doc_id_to_url", source_doc_id)
        dest_doc_url = self.core.call_success("doc_id_to_url", dest_doc_id)

        self.core.call_all(
            "page_move_by_url",
            source_doc_url, source_page_idx,
            dest_doc_url, dest_page_idx
        )

        self.core.call_success("transaction_simple", [
            ("upd", source_doc_id),
            ("upd", dest_doc_id),
        ])
        self.core.call_success("mainloop_quit_graceful")
        self.core.call_success("mainloop")

        return True



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import shutil

import openpaperwork_core

from . import util
from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        # if there are no doc_renderer loaded, nothing is displayed, which
        # may be fine.
        # (see paperwork-json)
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser('show', help=_(
            "Show the content of a document"
        ))
        p.add_argument('doc_id')
        p.add_argument(
            '--pages', '-p', required=False,
            help="Pages to show: 1,4 or 1-10 (default: all)"
        )

    def cmd_run(self, console, args):
        if args.command != 'show':
            return None

        doc_id = args.doc_id
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        if doc_url is None:
            return False
        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None or nb_pages <= 0:
            return False

        pages = util.parse_page_list(args)
        if pages is None:
            pages = range(0, nb_pages)

        renderers = []
        self.core.call_all("doc_renderer_get", renderers)
        assert len(renderers) > 0
        renderer = renderers[-1]

        header = _("Document id: %s") % doc_id
        self.core.call_success("print", header)
        self.core.call_success("print", "=" * len(header))

        doc_date = self.core.call_success("doc_get_date_by_id", doc_id)
        doc_date = self.core.call_success("i18n_date_short", doc_date)
        header = _("Document date: %s") % doc_date
        self.core.call_success("print", header)

        lines = renderer.get_doc_output(
            doc_id, doc_url, shutil.get_terminal_size()
        )
        for line in lines:
            self.core.call_success("print", line)
        self.core.call_success("print", "")

        for page_nb in pages:
            self.core.call_success("print", "")
            header = _("Page %d") % (page_nb + 1)
            self.core.call_success("print", header)
            self.core.call_success("print", ("-" * len(header)) + "\n")
            lines = renderer.get_page_output(
                doc_id, doc_url, page_nb, shutil.get_terminal_size()
            )
            for line in lines:
                self.core.call_success("print", line)
            self.core.call_success("print", "")

        self.core.call_success("print_flush")

        return {
            'document': renderer.get_doc_infos(doc_id, doc_url),
            'pages': {
                page_nb: renderer.get_page_infos(doc_id, doc_url, page_nb)
                for page_nb in pages
            },
        }



import logging

import openpaperwork_core

from . import BaseDocViewController


LOGGER = logging.getLogger(__name__)


class ZoomLayoutController(BaseDocViewController):
    def __init__(self, plugin):
        super().__init__(plugin)
        self.last_zoom = -1

    def _recompute_zoom(self):
        layout_name = self.plugin.layout_name
        spacing = self.plugin.page_layout.get_column_spacing()
        nb_columns = self.plugin.LAYOUTS[layout_name]
        max_columns = 0
        view_width = self.plugin.scroll.get_allocated_width()
        zoom = 1.0
        for page_idx in range(0, len(self.plugin.pages), nb_columns):
            pages = self.plugin.pages[page_idx:page_idx + nb_columns]
            pages_width = sum([p.get_full_size()[0] for p in pages])
            pages_width += (len(pages) * 30 * spacing) + 1
            zoom = min(zoom, view_width / pages_width)
            max_columns = max(max_columns, len(pages))

        if zoom == self.last_zoom:
            return
        self.last_zoom = zoom

        self.plugin.core.call_all("docview_set_zoom", zoom)
        for page in self.plugin.pages:
            page.set_zoom(zoom)

        if nb_columns > 1:
            layout = 'grid'
        else:
            layout = 'paged'
        self.plugin.core.call_all("on_layout_change", layout)

    def enter(self):
        super().enter()
        self._recompute_zoom()

    def docview_set_zoom(self, zoom):
        super().docview_set_zoom(zoom)
        if zoom == self.last_zoom:
            return
        self.plugin.docview_switch_controller('zoom', ZoomCustomController)

    def docview_set_layout(self, name):
        super().docview_set_layout(name)
        self._recompute_zoom()

    def on_page_size_obtained(self, page):
        super().on_page_size_obtained(page)
        self._recompute_zoom()

    def on_layout_size_allocate(self, layout):
        self._recompute_zoom()

    def doc_reload(self):
        self._recompute_zoom()


class ZoomCustomController(BaseDocViewController):
    def _reapply_zoom(self):
        zoom = self.plugin.zoom
        for page in self.plugin.pages:
            page.set_zoom(zoom)

    def enter(self):
        super().enter()
        self._reapply_zoom()

    def docview_set_zoom(self, zoom):
        super().docview_set_zoom(zoom)
        self._reapply_zoom()

    def docview_set_layout(self, name):
        super().docview_set_layout(name)
        self.plugin.docview_switch_controller('zoom', ZoomLayoutController)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['gtk_docview_controller']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
        ]

    def gtk_docview_get_controllers(self, out: dict, docview):
        out['zoom'] = ZoomLayoutController(docview)



import openpaperwork_core


PREVIEW_MAX_LINES = 25


class TextRenderer(object):
    def __init__(self, core):
        self.core = core
        self.parent = None

    def _get_page_text(self, doc_url, page_nb):
        out = []
        line_boxes = self.core.call_success(
            "page_get_boxes_by_url", doc_url, page_nb
        )
        if line_boxes is None:
            return out
        for line_box in line_boxes:
            line = line_box.content
            if line == "":
                continue
            out.append(line)
        return out

    def _rearrange_lines(self, lines, terminal_width):
        out = []
        for line in lines:
            new_line = ""
            iwords = line.split(" ")
            words = []
            for word in iwords:
                for p in range(0, len(word), terminal_width - 1):
                    words.append(word[p:p + terminal_width - 1])
            for word in words:
                if len(new_line) + len(word) + 1 >= terminal_width:
                    out.append(new_line)
                    new_line = ""
                new_line += " " + word
            if len(new_line.strip()) > 0:
                out.append(new_line)
        return [line.strip() for line in out]

    def get_preview_output(
                self, doc_id, doc_url, terminal_size=(80, 25),
                page_idx=0
            ):
        if self.parent is None:
            out = []
        else:
            out = self.parent.get_preview_output(
                doc_id, doc_url, terminal_size, page_idx
            )
        text = self._get_page_text(doc_url, page_idx)
        text = self._rearrange_lines(text, terminal_size[0])
        text = text[:PREVIEW_MAX_LINES]
        return out + text

    def get_doc_output(self, doc_id, doc_url, terminal_size=(80, 25)):
        if self.parent is None:
            out = []
        else:
            out = self.parent.get_doc_output(doc_id, doc_url, terminal_size)

        return out

    def get_page_output(
                self, doc_id, doc_url, page_nb, terminal_size=(80, 25)
            ):
        if self.parent is None:
            out = []
        else:
            out = self.parent.get_page_output(doc_id, doc_url, terminal_size)
        text = self._get_page_text(doc_url, page_nb)
        text = self._rearrange_lines(text, terminal_size[0])
        return out + text

    def get_doc_infos(self, doc_id, doc_url):
        out = {}
        if self.parent is not None:
            out = self.parent.get_doc_infos(doc_id, doc_url)
        out["doc_id"] = doc_id
        return out

    def get_page_infos(self, doc_id, doc_url, page_nb):
        out = {}
        if self.parent is not None:
            out = self.parent.get_page_infos(doc_id, doc_url, page_nb)
        out["text"] = self._get_page_text(doc_url, page_nb)
        return out


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def get_interfaces(self):
        return ['doc_renderer']

    def get_deps(self):
        return [
            {
                'interface': 'page_boxes',
                'defaults': [
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
        ]

    def doc_renderer_get(self, out):
        r = TextRenderer(self.core)
        if len(out) > 0:
            r.parent = out[-1]
        out.append(r)



import os
try:
    import pkg_resources
    PKG_RESOURCES_AVAILABLE = True
except Exception:
    PKG_RESOURCES_AVAILABLE = False
import shutil

import xdg.BaseDirectory
import xdg.DesktopEntry
import xdg.IconTheme

import openpaperwork_core

from .. import _


ICON_SIZES = [
    16, 22, 24, 30, 32, 36, 42, 48, 50, 64, 72, 96, 100, 128, 150, 160,
    192, 256, 512
]


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return []

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser('install', help=_(
            "Install Paperwork icons and shortcuts"
        ))
        p.add_argument(
            "--user", "-u",
            help=_("Install everything only for the current user"),
            action="store_true"
        )
        p.add_argument("--icon_base_dir", default="/usr/share/icons")
        p.add_argument("--data_base_dir", default="/usr/share")

    def _install(self, console, icondir, datadir):
        assert PKG_RESOURCES_AVAILABLE
        png_src_icon_pattern = "paperwork_{}.png"
        png_dst_icon_pattern = os.path.join(
            icondir, "hicolor", "{size}x{size}", "apps",
            "work.openpaper.Paperwork.png"
        )
        desktop_path = os.path.join(
            datadir, 'applications', 'work.openpaper.Paperwork.desktop'
        )
        appdata_path = os.path.join(
            datadir, "metainfo", "work.openpaper.Paperwork.metainfo.xml"
        )

        os.makedirs(os.path.dirname(desktop_path), exist_ok=True)

        to_copy = [
            (
                pkg_resources.resource_filename(
                    'paperwork_gtk.data',
                    png_src_icon_pattern.format(size)
                ),
                png_dst_icon_pattern.format(size=size),
            ) for size in ICON_SIZES
        ]
        to_copy.append(
            (
                pkg_resources.resource_filename(
                    'paperwork_gtk.data',
                    'work.openpaper.Paperwork.metainfo.xml',
                ),
                appdata_path
            )
        )
        for icon in ['paperwork.svg', 'paperwork_halo.svg']:
            src_icon = icon
            dst_icon = icon
            if icon == 'paperwork.svg':
                dst_icon = 'work.openpaper.Paperwork.svg'
            to_copy.append(
                (
                    pkg_resources.resource_filename(
                        'paperwork_gtk.data', src_icon
                    ),
                    os.path.join(
                        icondir, "hicolor", "scalable", "apps", dst_icon
                    )
                )
            )

        for (src, dst) in to_copy:
            console.print("Installing {} ...".format(dst))
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            shutil.copyfile(src, dst)

        console.print("Generating {} ...".format(desktop_path))
        entry = xdg.DesktopEntry.DesktopEntry(desktop_path)
        entry.set("GenericName", "Personal Document Manager")
        entry.set("Type", "Application")
        entry.set("Categories", "Office;Scanning;OCR;Archiving;GNOME;")
        entry.set("Terminal", "false")
        entry.set("Comment", "Grepping dead trees the easy way")
        # It's possible to add several actions to a single desktop file.
        # Ideally, we want a main "Open paperwork" and a secondary "Import
        # these files to paperwork". Unfortunately, the specification does
        # not allow specifying different MimeType= for the main and secondary
        # actions. So we always import (possibly 0) files.
        # https://specifications.freedesktop.org/desktop-entry-spec/1.5/ar01s11.html
        entry.set("Exec", "paperwork-gtk import %U")
        entry.set("Name", "Paperwork")
        entry.set("Icon", "work.openpaper.Paperwork")
        entry.set("Keywords", "document;scanner;ocr;")
        # PDF and all image formats supported by pillow
        entry.set(
            "MimeType",

            "application/pdf;"
            "image/bmp;"
            "image/gif;"
            "image/ico;"
            "image/icon;"
            "image/jp2;"
            "image/jpeg2000;"
            "image/jpeg;"
            "image/jpg;"
            "image/jpx;"
            "image/pjpeg;"
            "image/png;"
            "image/tiff;"
            "image/webp;"
            "image/x-bmp;"
            "image/x-MS-bmp;"
            "image/x-png;"
            "image/x-portable-bitmap;"
            "image/x-portable-graymap;"
            "image/x-portable-pixmap;"
            "image/x-tga;"
        )
        entry.validate()
        entry.write()
        console.print("Done")

    def cmd_run(self, console, args):
        if args.command != 'install':
            return None

        icon_base_dir = args.icon_base_dir
        data_base_dir = args.data_base_dir
        if args.user:
            icon_base_dir = xdg.IconTheme.icondirs[0]
            data_base_dir = xdg.BaseDirectory.xdg_data_dirs[0]

        self._install(console, icon_base_dir, data_base_dir)
        return True



import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['home']

    def get_deps(self):
        return [
            {
                'interface': 'app',
                'defaults': ['paperwork_backend.app'],
            },
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'icon',
                'defaults': ['paperwork_gtk.icon'],
            },
        ]

    def init(self, core):
        super().init(core)
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.home", "home.glade"
        )
        if widget_tree is None:
            return

        logo_pixbuf = self.core.call_success(
            "icon_get_pixbuf", "paperwork", 128
        )
        logo_widget = widget_tree.get_object("paperwork_logo")
        logo_widget.set_from_pixbuf(logo_pixbuf)

        logo_text = widget_tree.get_object("main_label")
        logo_text.set_text("Paperwork {}".format(
            self.core.call_success("app_get_version")
        ))

        self.core.call_all(
            "mainwindow_add", side="right", name="home", prio=100000,
            header=None,
            body=widget_tree.get_object("home")
        )



"""
Doc and page exporting are designed as pipelines.
UI build the pipelines and run it.

Data to export are represented as tree:
    Document set to export: data = set of (doc_id, doc_url))
    |-- Document to export: data = (doc_id, doc_url)
    |   |-- Page to export: data = (page_idx)
    |   |   |-- img_boxes: data = (Pillow image, boxes)
    |   |-- Page to export
    |
    |-- Document_to_export
    | (...)

When a pipes says it needs a given type in input, it means the content
of the tree must be extended up to this data type.

For example, if a pipe says it needs ExportDataType.PAGE as input, it must
get an ExportData of type ExportDataType.DOCUMENT_SET as input, but expanded
up to ExportDataType.PAGE:
    Document set to export
    |-- Document to export
    |   |-- Page to export
    |   |-- Page to export
    |
    |-- Document_to_export
    | (...)

It must *not* be expanded any further either (for instance, no img+boxes
(ExportDataType.IMG_BOXES) if the pipe says it expects a ExportDataType.PAGE
as input).

There are pipes dedicated to expanding unexpanded data (see docexport.generic).
"""
import abc
import enum

import openpaperwork_core


class ExportDataType(enum.Enum):
    DOCUMENT_SET = 0
    DOCUMENT = 1
    PAGE = 2
    IMG_BOXES = 3

    # the following ones are for final output only.
    # A list of paths (str) will be returned instead of an ExportData object.
    OUTPUT_URL_FILE = -1


class ExportData:
    def __init__(self, dtype, data):
        self.dtype = dtype
        self.data = data
        self._children = []
        self.expanded = False

    def clone(self):
        out = ExportData(self.dtype, self.data)
        out._children = [c.clone() for c in self._children]
        out.expanded = self.expanded
        return out

    def iter(self, dtype):
        if self.dtype == dtype:
            yield self
            return
        for c in self.get_children():
            for s in c.iter(dtype):
                yield (self, s)

    def get_children(self):
        assert self.expanded
        return self._children

    def set_children(self, children):
        self._children = children
        self.expanded = True

    @staticmethod
    def build_doc_set(docs):
        return ExportData(ExportDataType.DOCUMENT_SET, docs)

    @staticmethod
    def build_doc(doc_id, doc_url):
        root = ExportData.build_doc_set({(doc_id, doc_url)})
        doc = ExportData(ExportDataType.DOCUMENT, (doc_id, doc_url))
        root.set_children([doc])
        return root

    @staticmethod
    def build_page(doc_id, doc_url, page_idx):
        return ExportData.build_pages(doc_id, doc_url, [page_idx])

    @staticmethod
    def build_pages(doc_id, doc_url, page_indexes):
        root = ExportData.build_doc_set({(doc_id, doc_url)})
        doc = ExportData(ExportDataType.DOCUMENT, (doc_id, doc_url))
        pages = [
            ExportData(ExportDataType.PAGE, page_idx)
            for page_idx in page_indexes
        ]
        doc.set_children(pages)
        root.set_children([doc])
        return root


class AbstractExportPipe(abc.ABC):
    """
    Pipes are used by the frontend/user to build an export pipeline (a list of
    pipes to apply in a specific order).

    There are input pipes, taking either DOCUMENT_SET, DOCUMENT, or PAGES
    as input.
    There are output pipes, providing a file URL (FILE_URL) as output
    or a directory URL.
    And there are processing pipes (usually taking Pillow images and text boxes
    as input (IMG_BOXES)).

    Once the pipeline is defined, the frontend code can obtain promises
    from the pipes (`export_get_pipe_*()`), chain them together, and schedule
    them (see get_promise()).
    """

    def __init__(self, name, input_type, output_type):
        """
        Arguments:
          name -- name of the export pipe
          input_type -- accepted ExportDataType
          output_type -- ExportDataType
        """
        self.name = name
        self.input_type = input_type
        self.output_type = output_type
        self.can_change_quality = False
        self.can_change_page_format = False

        self.quality = 0.75
        self.page_format = (595.2755905511812, 841.8897637795276)  # A4

    def can_export_doc(self, doc_url):
        return False

    def can_export_page(self, doc_url, page_idx):
        return False

    @abc.abstractmethod
    def export(self, input_data, result="final", target_file_url=None):
        ...

    def get_promise(self, result='final', target_file_url=None):
        """
        Returns a promise.
        Beware that the promise will modify the ExportData object given as
        input. If you want to preserve your copy, please use
        ExportData.clone().

        Arguments:
        - result: either 'preview' (only one page) or 'final' (all pages)
        - target_file_url: Where the final pipe should write the file
          (None = temporary file)

        Returns:
        - A promise, that expect ExportData object as input and will
          return either an ExportData (if it must be chained to another pipe)
          or a list of str (list of paths)
        """
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.export, kwargs={
                "result": result,
                "target_file_url": target_file_url,
            }
        )

    def get_estimated_size_factor(self, input_data):
        """
        Return the factor to apply to the preview size to get an estimation
        of the final result size.

        Arguments:
          input_data -- ExportData. Won't be modified
        """
        return 1

    def set_quality(self, quality):
        """
        Allow to define an output quality (between 0 and 100).
        Check `can_change_quality` before calling this method.
        """
        assert self.can_change_quality
        self.quality = quality

    def set_page_format(self, page_format):
        """
        Allow to define the expected output page format.
        Check `can_change_quality` before calling this method.

        Arguments:
         page_format: tuple (width, height), in points (1 point == 1/72.0 inch)
        """
        assert self.can_change_page_format
        self.page_format = page_format

    def get_output_mime(self):
        """
        If the pipe outputs a file, specifies its mime type and
        a list of possible file extensions.
        None if it doesn't outputs a file.
        """
        return None

    @abc.abstractmethod
    def __str__(self):
        ...


class ExportDataTransformedImgBoxes(ExportData):
    """
    Page images takes a lot of memory --> we only generate them when actually
    requested.
    """
    def __init__(self, pipe, original_page):
        super().__init__(ExportDataType.PAGE, original_page.data)
        self.expanded = True
        self.pipe = pipe
        self.original_page = original_page

    def get_children(self):
        children = self.original_page.get_children()
        for img_boxes in children:
            (img, boxes) = img_boxes.data
            img = self.pipe.transform(img)
            yield ExportData(ExportDataType.IMG_BOXES, (img, boxes))


class AbstractSimpleTransformExportPipe(AbstractExportPipe):
    """
    Base template class for page image to page image transformations.
    """
    def __init__(self, core, name):
        super().__init__(
            name=name,
            input_type=ExportDataType.IMG_BOXES,
            output_type=ExportDataType.IMG_BOXES
        )
        self.core = core

    def transform(self, img):
        # sub-classes must implement it
        assert False

    def export(self, input_data, result='final', target_file_url=None):
        assert input_data.dtype == ExportDataType.DOCUMENT_SET

        docs = input_data.iter(ExportDataType.DOCUMENT)
        docs = list(docs)

        # replace the document page list by objects that will
        # generate their children (img+boxes) on-the-fly.
        for (doc_set, doc) in docs:
            assert doc.expanded
            doc.set_children([
                ExportDataTransformedImgBoxes(self, page)
                for page in doc.get_children()
            ])

        return input_data


class AbstractExportPipePlugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.pipes = []

    def get_interfaces(self):
        return ['export_pipes']

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def export_get_pipe_by_name(self, name):
        for pipe in self.pipes:
            if pipe.name == name:
                return pipe

    def export_get_pipes_by_input(self, out: list, input_type):
        for pipe in self.pipes:
            if input_type == pipe.input_type:
                out.append(pipe)

    def export_get_pipes_by_doc_urls(self, out: list, doc_urls):
        for pipe in self.pipes:
            if pipe.input_type != ExportDataType.DOCUMENT:
                continue
            for doc_url in doc_urls:
                if not pipe.can_export_doc(doc_url):
                    break
            else:
                out.append(pipe)

    def export_get_pipes_by_doc_url(self, out: list, doc_url):
        return self.export_get_pipes_by_doc_urls(out, [doc_url])

    def export_get_pipes_by_page(self, out: list, doc_url, page_nb):
        for pipe in self.pipes:
            if pipe.input_type != ExportDataType.PAGE:
                continue
            if not pipe.can_export_page(doc_url, page_nb):
                continue
            out.append(pipe)



from .. import _


def ask_confirmation(
            console, question,
            default_interactive='n', default_non_interactive='n'
        ):
    if default_interactive == 'y':
        yesno = _("Y/n")
    else:
        yesno = _("y/N")
    if default_interactive == 'y':
        default_interactive = yesno[0].lower()  # l10n
    else:
        default_interactive = yesno[-1].lower()  # l10n
    if default_non_interactive == 'y':
        default_non_interactive = yesno[0].lower()  # l10n
    else:
        default_non_interactive = yesno[-1].lower()  # l10n

    try:
        reply = console.input(f"{question} [{yesno}] ")
    except KeyboardInterrupt:
        return 'n'
    if reply is None:
        reply = default_non_interactive
    else:
        reply = reply.strip().lower()
        if reply == "":
            reply = default_interactive

    return 'y' if reply == yesno[0].lower() else 'n'



import datetime
import logging
import os
import re

import openpaperwork_core
import openpaperwork_core.beacon
import openpaperwork_core.promise

from .. import _version


LOGGER = logging.getLogger(__name__)

UPDATE_CHECK_INTERVAL = datetime.timedelta(days=7)
UPDATE_PATH = "/beacon/latest"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.periodic = None
        self.http = None

    def get_interfaces(self):
        return ["update_detection"]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'http_json',
                'defaults': ['openpaperwork_core.http'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def _register_config(self, core):
        setting = self.core.call_success(
            "config_build_simple", "update",
            "enabled", lambda: False
        )
        self.core.call_all(
            "config_register", "check_for_update", setting
        )
        setting = self.core.call_success(
            "config_build_simple", "update",
            "last_update_found", lambda: _version.version
        )
        self.core.call_all(
            "config_register", "last_update_found", setting
        )

    def _parse_version(self, version):
        version = re.split('[^0-9]', version)[:3]
        version = tuple([
            int(x.strip())
            for x in version
            if x.strip() != ''
        ])
        return version

    def update_check(self):
        LOGGER.info("Looking for updates...")

        def on_success(update_data, core):
            latest_version = update_data['paperwork'][os.name]
            LOGGER.info("Version advertised: %s", latest_version)
            self.core.call_all(
                "config_put", "last_update_found", latest_version
            )
            self.core.call_all("config_save")
            self.update_compare()

        promise = openpaperwork_core.promise.Promise(self.core, lambda: "")
        promise = promise.then(self.http.get_request_promise(UPDATE_PATH))
        promise = promise.then(on_success, self.core)
        promise = promise.catch(self._on_upd_check_error)
        promise.schedule()

    def _on_upd_check_error(self, exc):
        LOGGER.warning("Failed to look for update", exc_info=exc)

    def update_compare(self):
        remote_version = self.core.call_success(
            "config_get", "last_update_found"
        )
        remote_version = self._parse_version(remote_version)
        LOGGER.info("Remote version: %s", remote_version)
        local_version = self._parse_version(_version.version)
        LOGGER.info("Current version: %s", local_version)
        if remote_version > local_version:
            self.core.call_all(
                "on_update_detected", local_version, remote_version
            )

    def init(self, core):
        super().init(core)
        self.periodic = openpaperwork_core.beacon.PeriodicTask(
            "update",
            UPDATE_CHECK_INTERVAL,
            self.update_check,
            self.update_compare
        )
        self.http = self.core.call_success("http_json_get_client", "update")
        self._register_config(core)
        self.periodic.register_config(core)

        if self.core.call_success("config_get", "check_for_update"):
            self.core.call_all("mainloop_schedule", self.periodic.do, core)



import logging

import PIL
import PIL.Image

from .. import (
    PluginBase,
    promise
)


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    FILE_EXTENSIONS = [
        "bmp",
        "gif",
        "jpeg",
        "jpg",
        "png",
        "tif",
        "tiff",
    ]

    def get_interfaces(self):
        return [
            'page_img_size',
            'pillow',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python']
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                'interface': 'pillow_util',
                'defaults': ['openpaperwork_core.pillow.util'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def _check_is_img(self, file_url):
        return file_url.rsplit(".", 1)[-1].lower() in self.FILE_EXTENSIONS

    def url_to_img_size(self, file_url):
        if not self._check_is_img(file_url):
            return None

        task = "url_to_img_size({})".format(file_url)
        size = (0, 0)
        self.core.call_all("on_perfcheck_start", task)
        try:
            with self.core.call_success("fs_open", file_url, mode='rb') as fd:
                img = PIL.Image.open(fd)
                size = img.size
            return size
        except PIL.Image.DecompressionBombError as exc:
            LOGGER.warning("Image %s is too big", file_url, exc_info=exc)
            return self.core.call_success(
                "pillow_get_error", "too_big"
            ).size
        finally:
            self.core.call_all("on_perfcheck_stop", task, size=size)

    def url_to_img_size_promise(self, file_url):
        if not self._check_is_img(file_url):
            return None

        return promise.ThreadedPromise(
            self.core, self.url_to_img_size, args=(file_url,)
        )

    def url_to_pillow(self, file_url):
        if not self._check_is_img(file_url):
            return None

        task = "url_to_pillow({})".format(file_url)
        size = (0, 0)
        self.core.call_all("on_perfcheck_start", task)
        try:
            with self.core.call_success("fs_open", file_url, mode='rb') as fd:
                img = PIL.Image.open(fd)
                img.load()
                self.core.call_all("on_objref_track", img)
                size = img.size
            return img
        except PIL.Image.DecompressionBombError as exc:
            LOGGER.warning("Image %s is too big", file_url, exc_info=exc)
            return self.core.call_success(
                "pillow_get_error", "too_big"
            )
        finally:
            self.core.call_all("on_perfcheck_stop", task, size=size)

    def url_to_pillow_promise(self, file_url):
        return promise.ThreadedPromise(
            self.core, self.url_to_pillow, args=(file_url,)
        )

    def pillow_to_url(self, img, file_url, format=None, quality=0.75):
        expected_format = file_url.rsplit(".", 1)[-1].upper()
        if expected_format == "JPG":
            expected_format = "JPEG"
        if expected_format == "TIF":
            expected_format = "TIFF"
        if format is None:
            format = expected_format
        if format != expected_format:
            LOGGER.warning(
                "File %s written with format %s. Expected %s",
                file_url, format, expected_format
            )
        if format != 'PNG':
            # drop the alpha channel if there is one
            img = img.convert("RGB")
        with self.core.call_success("fs_open", file_url, mode='wb') as fd:
            return img.save(
                fd, format=format, optimize=True, quality=int(quality * 100)
            )



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise
import openpaperwork_gtk.deps

from .... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.busy = False

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_scan_buttons_popover_sources',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_scan_buttons_popover',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageadd.source_popover'
                ],
            },
            {
                'interface': 'i18n_scanner',
                'defaults': ['paperwork_backend.i18n.scanner'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
            {
                'interface': 'scan2doc',
                'defaults': ['paperwork_backend.docscan.scan2doc'],
            },
        ]

    def init(self, core):
        super().init(core)

        opt = self.core.call_success(
            "config_build_simple", "pageadd", "scanner_sources", lambda: []
        )
        self.core.call_all("config_register", "pageadd_sources", opt)

        self.core.call_all(
            "config_add_observer", "scanner_dev_id", self._update_sources
        )

        self.core.call_all(
            "mainloop_schedule", self.core.call_all,
            "pageadd_sources_refresh"
        )

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def _update_sources(self):
        def get_sources(dev=None):
            if dev is None:
                source_names = []
            else:
                sources = dev.dev.get_children()
                source_names = []
                for src in sources:
                    source_names.append(src.get_name())
                sources = None
            LOGGER.info("Scanner sources: %s", source_names)
            return source_names

        def store_sources(sources):
            self.core.call_all("config_put", "pageadd_sources", list(sources))

        promise = self.core.call_success("scan_get_scanner_promise")
        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, get_sources
        ))
        promise = promise.then(store_sources)
        promise = promise.then(self.core.call_all, "pageadd_sources_refresh")
        self.core.call_success("scan_schedule", promise)

    def pageadd_get_sources(self, out: list):
        sources = self.core.call_success("config_get", "pageadd_sources")
        for source in list(sources):
            widget_tree = self.core.call_success(
                "gtk_load_widget_tree",
                "paperwork_gtk.mainwindow.docview.pageadd", "source_box.glade"
            )
            source_txt = self.core.call_success(
                "i18n_scanner_source", source
            )
            source_txt = _("Scan from %s") % source_txt

            img = "view-paged-symbolic"
            if "flatbed" in source:
                img = "document-new-symbolic"

            widget_tree.get_object("source_image").set_from_icon_name(
                img, Gtk.IconSize.SMALL_TOOLBAR
            )
            widget_tree.get_object("source_name").set_text(source_txt)

            out.append(
                (
                    widget_tree.get_object("source_selector"),
                    source_txt, source, self._on_scan
                )
            )

    def _on_scan(self, doc_id, doc_url, source_id):
        LOGGER.info("Scanning from %s", source_id)

        if doc_id is None or doc_url is None:
            (doc_id, doc_url) = self.core.call_success("get_new_doc")
            self.core.call_all("doc_open", doc_id, doc_url)

        self.core.call_all("on_busy")
        self.core.call_all("pageadd_busy_add")
        self.busy = True

        promise = self.core.call_success(
            "scan2doc_promise", doc_id=doc_id, doc_url=doc_url,
            source_id=source_id
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_all, "pageadd_busy_remove"
        )
        self.core.call_success("scan_schedule", promise)

    def on_scan2doc_page_scanned(self, scan_id, doc_id, doc_url, nb_pages):
        LOGGER.info(
            "Page scanned: %s p%d (scan id = %d)", doc_id, nb_pages, scan_id
        )
        self.core.call_all("doc_reload", doc_id, doc_url)

    def on_scan_started(self, scan_id):
        if not self.busy:
            return
        self.core.call_all("on_idle")
        self.busy = False



import random

import openpaperwork_core

from . import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'update_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'icon',
                'defaults': ['paperwork_gtk.icon'],
            },
            {
                'interface': 'notifications',
                'defaults': ['paperwork_gtk.notifications.notify'],
            },
            {
                'interface': 'update_detection',
                'defaults': ['paperwork_backend.beacon.update'],
            },
        ]

    def on_update_detected(self, local_version, remote_version):
        random_dumbnesses = [
            _("Now with 10% more freedom in it !"),
            _("Buy it now and get a 100% discount !"),
            _("New features and bugs available !"),
            _("New taste !"),
            _("We replaced your old bugs with new bugs. Enjoy."),
            _("Smarter, Better, Stronger"),
            # Linus Torvalds citation, look it up :)
            _("It's better when it's free."),
        ]

        notification = self.core.call_success(
            "get_notification_builder",
            _("A new version of Paperwork is available: {new_version}").format(
                new_version=".".join([str(x) for x in remote_version])
            )
        )
        if notification is None:
            return
        icon = self.core.call_success("icon_get_pixbuf", "paperwork", 32)
        notification.set_message(
            random.choice(random_dumbnesses)
        ).set_image_from_pixbuf(
            icon
        ).show()



try:
    import fabulous.color
    FABULOUS_AVAILABLE = True
except (ValueError, ImportError):
    FABULOUS_AVAILABLE = False

import openpaperwork_core

import rich.text


def color_labels(core, labels):
    labels = [
        (label, core.call_success("label_color_to_rgb", color))
        for (label, color) in labels
    ]
    labels = [
        (label, (
            int(color[0] * 0xFF),
            int(color[1] * 0xFF),
            int(color[2] * 0xFF)
        ))
        for (label, color) in labels
    ]

    for (label, bg_color) in labels:
        fg_color = core.call_success(
            "label_get_foreground_color",
            (
                bg_color[0] / 0xFF,
                bg_color[1] / 0xFF,
                bg_color[2] / 0xFF,
            )
        )
        fg_color = (fg_color[0] * 0xFF, fg_color[1] * 0xFF, fg_color[2] * 0xFF)
        l_label = len(label)
        label = fabulous.color.fg256(fg_color, label)
        label = fabulous.color.bg256(bg_color, label)
        yield (l_label, str(label))


class LabelsRenderer(object):
    def __init__(self, core):
        self.core = core
        self.parent = None

    def _get_labels(self, doc_url):
        labels = set()
        self.core.call_all("doc_get_labels_by_url", labels, doc_url)
        return labels

    def _rearrange_labels(self, labels, terminal_width):
        out = []
        line = ""
        for (len_label, label) in labels:
            if len(line) + len_label + 1 >= terminal_width:
                out.append(line)
                line = ""
            line += " " + label
        if len(line.strip()) > 0:
            out.append(line)
        return [line.strip() for line in out]

    def get_preview_output(
                self, doc_id, doc_url, terminal_size=(80, 25),
                page_idx=0
            ):
        out = []
        if self.parent is not None:
            out = self.parent.get_preview_output(
                doc_id, doc_url, terminal_size, page_idx
            )
        if not FABULOUS_AVAILABLE:
            return out
        if page_idx != 0:
            return out

        labels = self._get_labels(doc_url)
        labels = color_labels(self.core, labels)
        labels = self._rearrange_labels(labels, terminal_size[0])
        return labels + out

    def get_doc_output(self, doc_id, doc_url, terminal_size=(80, 25)):
        out = []
        if self.parent is not None:
            out = self.parent.get_doc_output(
                doc_id, doc_url, terminal_size
            )
        if not FABULOUS_AVAILABLE:
            return out

        labels = self._get_labels(doc_url)
        labels = color_labels(self.core, labels)
        labels = self._rearrange_labels(labels, terminal_size[0])
        return labels + out

    def get_page_output(
                self, doc_id, doc_url, page_nb, terminal_size=(80, 25)
            ):
        if self.parent is not None:
            return self.parent.get_page_output(
                doc_id, doc_url, page_nb, terminal_size
            )
        return []

    def get_doc_infos(self, doc_id, doc_url):
        out = {}
        if self.parent is not None:
            out = self.parent.get_doc_infos(doc_id, doc_url)
        out['labels'] = []
        for (label, color) in self._get_labels(doc_url):
            out['labels'].append(
                {
                    'label': label,
                    'color': color,
                }
            )
        return out

    def get_page_infos(self, doc_id, doc_url, page_nb):
        if self.parent is not None:
            return self.parent.get_page_infos(doc_id, doc_url, page_nb)
        return {}


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def get_interfaces(self):
        return ['doc_renderer']

    def get_deps(self):
        return [
            {
                'interface': 'page_boxes',
                'defaults': ['paperwork_backend.model.labels'],
            },
        ]

    def format_labels(self, labels, separator='\n'):
        if FABULOUS_AVAILABLE:
            labels = color_labels(self.core, labels)
            labels = [label for (l_label, label) in labels]
            labels = separator.join(labels)
            labels = rich.text.Text.from_ansi(labels)
        else:
            labels = [label for (label, color) in labels]
            labels = separator.join(labels)
        return labels

    def doc_renderer_get(self, out):
        r = LabelsRenderer(self.core)
        if len(out) > 0:
            r.parent = out[-1]
        out.append(r)



import datetime
import time

import openpaperwork_core

from . import workdir


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def __init__(self):
        super().__init__()
        # expected in self.docs:
        # [
        #   {
        #     'id': 'some_id',
        #     'url': 'file:///some_work_dir/some_id',
        #     'mtime': 12345,  # unix timestamp
        #     'labels': [(label_name, label_color), ...]
        #     'hash': 12345,  # optional
        #     'text': "pouet",  # optional
        #     'page_boxes: [  # optional
        #       [LineBox, LineBox, ...],  # page 0
        #       [LineBox, LineBox, ...],  # page 1
        #       (...)
        #     ],
        #     'page_imgs': [  # optional
        #       (img_url, PIL.Image),  # page 0
        #       (img_url, PIL.Image),  # page 1
        #     ],
        #     'page_mtimes': [  # optional
        #       (img_url, mtime),  # page 0
        #       (img_url, mtime),  # page 1
        #     ],
        #     'page_hashes': [  # optional
        #       (img_url, hash),  # page 0
        #       (img_url, hash),  # page 1
        #     ],
        #     'page_paper_sizes': [  # optional
        #       (img_url, hash),  # page 0
        #       (img_url, hash),  # page 1
        #     ],
        #   },
        #   (...)
        # ]
        self.docs = []
        self.new_doc_idx = 0

    def get_interfaces(self):
        return [
            "doc_hash",
            "doc_labels",
            "doc_text",
            "doc_type",
            "document_storage",
            "page_boxes",
            "page_paper",
            "pillow",
        ]

    def get_deps(self):
        return []

    def storage_get_id(self):
        return "file:///home/jflesch/papers"

    def storage_get_all_docs(self, out: list, only_valid=True):
        out += [
            (doc['id'], doc['url'])
            for doc in self.docs
        ]

    def doc_id_to_url(self, doc_id, existing=True):
        for doc in self.docs:
            if doc['id'] == doc_id:
                return doc['url']
        return None

    def doc_url_to_id(self, doc_url):
        for doc in self.docs:
            if doc['url'] == doc_url:
                return doc['id']
        return None

    def is_doc(self, doc_url):
        return True

    def doc_get_hash_by_url(self, doc_url):
        for doc in self.docs:
            if doc['url'] == doc_url:
                if 'hash' in doc:
                    return doc['hash']

    def doc_get_mtime_by_url(self, doc_url):
        for doc in self.docs:
            if doc['url'] == doc_url:
                return doc['mtime']

    def doc_get_nb_pages_by_url(self, doc_url):
        for doc in self.docs:
            if doc['url'] == doc_url:
                l_boxes = len(doc['page_boxes']) if 'page_boxes' in doc else 0
                l_imgs = len(doc['page_imgs']) if 'page_imgs' in doc else 0
                l_mtimes = (
                    len(doc['page_mtimes']) if 'page_mtimes' in doc else 0
                )
                l_hashes = (
                    len(doc['page_hashes']) if 'page_hashes' in doc else 0
                )
                l_paper_sizes = (
                    len(doc['page_paper_sizes'])
                    if 'page_paper_sizes' in doc else 0
                )
                return max(l_boxes, l_imgs, l_mtimes, l_hashes, l_paper_sizes)
        return None

    def doc_get_text_by_url(self, out: list, doc_url):
        for doc in self.docs:
            if doc['url'] == doc_url:
                out.append(doc['text'])

    def doc_get_labels_by_url(self, out: set, doc_url):
        for doc in self.docs:
            if doc['url'] == doc_url:
                out.update(doc['labels'])

    def doc_has_labels_by_url(self, doc_url):
        for doc in self.docs:
            if doc['url'] == doc_url:
                return True if len(doc["labels"]) > 0 else None
        return None

    def doc_add_label_by_url(self, doc_url, label, color=None):
        if color is None:
            all_labels = set()
            self.labels_get_all(all_labels)

            for (label_name, c) in all_labels:
                if label_name == label:
                    color = c
                    break
            else:
                raise Exception(
                    "label {} provided without color,"
                    " but label is unknown".format(label)
                )

        for doc in self.docs:
            if doc['url'] == doc_url:
                doc['labels'].add((label, color))

        return True

    def page_has_text_by_url(self, doc_url, page_idx):
        for doc in self.docs:
            if doc['url'] == doc_url:
                if page_idx >= len(doc['page_boxes']):
                    return None
                return True
        return None

    def page_get_boxes_by_url(self, doc_url, page_idx):
        for doc in self.docs:
            if doc['url'] == doc_url:
                if page_idx >= len(doc['page_boxes']):
                    return None
                return doc['page_boxes'][page_idx]
        return None

    def page_set_boxes_by_url(self, doc_url, page_idx, boxes):
        for doc in self.docs:
            if doc['url'] == doc_url:
                if page_idx >= len(doc['page_boxes']):
                    missing = page_idx + 1 - len(doc['page_boxes'])
                    assert missing >= 1
                    doc['page_boxes'] += ([None] * missing)
                doc['page_boxes'][page_idx] = boxes

                text = ""
                for page_boxes in doc['page_boxes']:
                    if text != "":
                        text += "\n\n"
                    if page_boxes is None:
                        continue
                    for line_boxes in page_boxes:
                        text += line_boxes.content + "\n"
                doc['text'] = text
        return True

    def page_get_img_url(self, doc_url, page_idx, write=False, **kwargs):
        for doc in self.docs:
            if doc['url'] == doc_url:
                for k in [
                            'page_imgs', 'page_mtimes', 'page_hashes',
                            'page_sizes'
                        ]:
                    if k in doc:
                        if page_idx >= len(doc[k]):
                            return None
                        return doc[k][page_idx][0]

                if write:
                    return "file:///some_doc/new_page.jpeg"
                else:
                    return None
        return None

    def url_to_pillow(self, img_url):
        for doc in self.docs:
            if 'page_imgs' not in doc:
                continue
            for (page_img_url, img) in doc['page_imgs']:
                if page_img_url == img_url:
                    return img
        return None

    def labels_get_all(self, out: set):
        for doc in self.docs:
            out.update(doc['labels'])

    def storage_get_new_doc(self):
        self.new_doc_idx += 1
        doc = {
            'url': 'file:///some_work_dir/{}'.format(self.new_doc_idx),
            'id': str(self.new_doc_idx),
            'mtime': time.time(),
            'labels': [],
        }
        self.docs.append(doc)
        return (doc['id'], doc['url'])

    def doc_get_date_by_id(self, doc_id):
        # Doc id is expected to have this format:
        # YYYYMMDD_hhmm_ss_NN_something_else
        doc_id = doc_id.split("_", 3)
        doc_id = "_".join(doc_id[:3])
        try:
            return datetime.datetime.strptime(doc_id, workdir.DOCNAME_FORMAT)
        except ValueError:
            return None

    def storage_delete_doc_id(self, doc_id):
        for (idx, doc) in enumerate(self.docs[:]):
            if doc['id'] == doc_id:
                self.docs.pop(idx)
                return True

    def page_delete(self, doc_url, page_idx):
        raise NotImplementedError()

    def page_get_mtime_by_url(self, doc_url, page_idx):
        for doc in self.docs:
            if doc['url'] != doc_url:
                continue
            if 'page_mtimes' not in doc:
                continue
            return doc['page_mtimes'][page_idx][1]

    def page_get_hash_by_url(self, doc_url, page_idx):
        for doc in self.docs:
            if doc['url'] != doc_url:
                continue
            if 'page_hashes' not in doc:
                continue
            return doc['page_hashes'][page_idx][1]

    def page_get_paper_size_by_url(self, doc_url, page_idx):
        for doc in self.docs:
            if doc['url'] != doc_url:
                continue
            if 'page_paper_sizes' not in doc:
                continue
            return doc['page_paper_sizes'][page_idx][1]



import logging

import PIL
import PIL.Image
import pyocr
import pyocr.builders

import openpaperwork_core

from ... import (_, sync)


LOGGER = logging.getLogger(__name__)
ID = "orientation_guesser"


class OrientationTransaction(sync.BaseTransaction):
    def __init__(self, plugin, sync, total_expected=-1):
        super().__init__(plugin.core, total_expected)
        self.priority = plugin.PRIORITY

        self.plugin = plugin
        self.sync = sync

        # for each document, we need to track on which pages we have already
        # guessed the orientation and on which page we didn't yet.
        self.page_tracker = self.core.call_success("page_tracker_get", ID)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()

    def _guess_page_orientation(
            self, doc_id, doc_url, page_idx, page_nb, total_pages):
        if self.core.call_success(
                    "page_has_text_by_url", doc_url, page_idx
                ):
            self.notify_progress(
                ID,
                _(
                    "Document {doc_id} p{page_idx} has already some text."
                    " Not guessing page orientation."
                ).format(doc_id=doc_id, page_idx=(page_idx + 1)),
                page_nb=page_nb, total_pages=total_pages
            )
            return

        self.notify_progress(
            ID,
            _(
                "Guessing orientation of"
                " document {doc_id} p{page_idx}"
            ).format(doc_id=doc_id, page_idx=(page_idx + 1)),
            page_nb=page_nb, total_pages=total_pages
        )
        self.plugin.guess_page_orientation_by_url(doc_url, page_idx)

    def _guess_new_page_orientations(self, doc_id):
        doc_url = self.core.call_success("doc_id_to_url", doc_id)

        need_end_notification = False
        modified_pages = list(self.page_tracker.find_changes(doc_id, doc_url))
        for (page_nb, (change, page_idx)) in enumerate(modified_pages):
            # Guess page orientation on new pages, but only if we are
            # not synchronizing with the work directory
            if not self.sync and change == 'new':
                self._guess_page_orientation(
                    doc_id, doc_url, page_idx, page_nb, len(modified_pages)
                )
                need_end_notification = True
            self.page_tracker.ack_page(doc_id, doc_url, page_idx)

        if need_end_notification:
            self.notify_progress(
                ID, _("Guessing page orientation"),
                page_nb=len(modified_pages), total_pages=len(modified_pages)
            )

    def add_doc(self, doc_id):
        self._guess_new_page_orientations(doc_id)
        super().add_doc(doc_id)

    def upd_doc(self, doc_id):
        self._guess_new_page_orientations(doc_id)
        super().upd_doc(doc_id)

    def del_doc(self, doc_id):
        self.page_tracker.delete_doc(doc_id)
        super().del_doc(doc_id)

    def cancel(self):
        self.page_tracker.cancel()
        self.notify_done(ID)

    def commit(self):
        self.page_tracker.commit()
        self.notify_done(ID)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 2000

    def get_interfaces(self):
        return [
            "orientation_guesser",
            "syncable",  # actually satisfied by the plugin 'doctracker'
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_tracking',
                'defaults': ['paperwork_backend.doctracker']
            },
            {
                'interface': 'ocr_settings',
                'defaults': ['paperwork_backend.pyocr'],
            },
            {
                'interface': 'page_tracking',
                'defaults': ['paperwork_backend.pagetracker'],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
        ]

    def init(self, core):
        super().init(core)
        self.core.call_all(
            "doc_tracker_register", ID,
            lambda sync, total_expected=-1: OrientationTransaction(
                self, sync, total_expected
            )
        )

    def guess_page_orientation_by_url(self, doc_url, page_idx):
        if self.core.call_success("ocr_is_enabled") is None:
            LOGGER.info("OCR is disabled")
            return

        LOGGER.info(
            "Using OCR to guess orientation of page %d of %s",
            page_idx, doc_url
        )

        doc_id = self.core.call_success("doc_url_to_id", doc_url)

        if doc_id is not None:
            self.core.call_one(
                "mainloop_schedule", self.core.call_all,
                "on_page_modification_start", doc_id, page_idx
            )
        try:
            page_img_url = self.core.call_success(
                "page_get_img_url", doc_url, page_idx
            )

            for ocr_tool in pyocr.get_available_tools():
                LOGGER.info(
                    "Orientation guessing: Will use tool '%s'",
                    ocr_tool.get_name()
                )
                if ocr_tool.can_detect_orientation():
                    break
                LOGGER.warning(
                    "Orientation guessing: Tool '%s' cannot detect"
                    " orientation",
                    ocr_tool.get_name()
                )
            else:
                LOGGER.warning(
                    "Orientation guessing: No tool found able to detect"
                    " orientation"
                )
                return None

            ocr_langs = self.core.call_success("ocr_get_active_langs")

            img = self.core.call_success("url_to_pillow", page_img_url)

            try:
                r = ocr_tool.detect_orientation(img, lang="+".join(ocr_langs))
            except pyocr.PyocrException as exc:
                LOGGER.warning(
                    "Orientation guessing: Failed to guess orientation",
                    exc_info=exc
                )
                return None

            angle = r['angle']
            if angle == 0:
                return 0

            transform_enum = getattr(PIL.Image, 'Transpose', PIL.Image)
            t_angle = {
                90: transform_enum.ROTATE_90,
                180: transform_enum.ROTATE_180,
                270: transform_enum.ROTATE_270,
            }[angle]
            img = img.transpose(t_angle)

            page_img_url = self.core.call_success(
                "page_get_img_url", doc_url, page_idx, write=True
            )
            self.core.call_success("pillow_to_url", img, page_img_url)
        finally:
            if doc_id is not None:
                self.core.call_one(
                    "mainloop_schedule", self.core.call_all,
                    "on_page_modification_end", doc_id, page_idx
                )

        return angle



import logging

import openpaperwork_core

from . import BaseDocViewController


LOGGER = logging.getLogger(__name__)


class AutoScrollingController(BaseDocViewController):
    def __init__(self, plugin):
        super().__init__(plugin)
        self.handler = None
        self.enter()

    def enter(self):
        if self.handler is not None:
            return
        self.handler = self.plugin.core.call_success(
            "gesture_enable_autoscrolling", self.plugin.scroll
        )

    def exit(self):
        if self.handler is None:
            return
        self.handler.disable()
        self.handler = None

    def on_close(self):
        self.exit()


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['gtk_docview_controller']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
            {
                'interface': 'gtk_gesture_autoscrolling',
                'defaults': ['openpaperwork_gtk.gesturee.autoscrolling'],
            },
        ]

    def gtk_docview_get_controllers(self, out: dict, docview):
        out['autoscrolling'] = AutoScrollingController(docview)



import logging

LOGGER = logging.getLogger(__name__)

try:
    import gi
    gi.require_version('Notify', '0.7')
    from gi.repository import Notify
    NOTIFY_AVAILABLE = True
except (ImportError, ValueError) as exc:
    LOGGER.error("Failed to import Notify", exc_info=exc)
    NOTIFY_AVAILABLE = False


import openpaperwork_core  # noqa E402
import openpaperwork_gtk.deps  # noqa E402


class NotifyBuilder(object):
    def __init__(self, title):
        self.title = title
        self.msg = None
        self.icon = None
        self.pixbuf = None
        self.actions = []
        self.notification = None

    def set_message(self, message):
        self.msg = message
        return self

    def set_icon(self, icon):
        self.icon = icon
        return self

    def add_action(self, action_id, label, callback, *args, **kwargs):
        self.actions.append((action_id, label, callback, args, kwargs))
        return self

    def set_image_from_pixbuf(self, pixbuf):
        self.pixbuf = pixbuf
        return self

    @staticmethod
    def _call_callback(notification, action, args):
        (callback, args, kwargs) = args
        return callback(*args, **kwargs)

    def show(self):
        self.notification = Notify.Notification.new(
            self.title, self.msg, self.icon
        )
        if self.pixbuf is not None:
            self.notification.set_image_from_pixbuf(self.pixbuf)
        for (action_id, label, callback, args, kwargs) in self.actions:
            self.notification.add_action(
                action_id, label,
                self._call_callback, (callback, args, kwargs)
            )
        self.notification.show()


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def __init__(self):
        # WORKAROUND(Jflesch): Keep a reference to the notifications.
        # Otherwise we never get the action from the user.
        self.notification_refs = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'notifications',
        ]

    def get_deps(self):
        return [
        ]

    def init(self, core):
        super().init(core)
        if NOTIFY_AVAILABLE:
            Notify.init("Paperwork")

    def chkdeps(self, out: dict):
        if not NOTIFY_AVAILABLE:
            out['notify'].update(openpaperwork_gtk.deps.NOTIFY)

    def get_notification_builder(self, title, need_actions=False):
        caps = []
        if NOTIFY_AVAILABLE:
            caps = Notify.get_server_caps()
            if len(caps) <= 0:
                return None
        if not need_actions or "actions" in caps:
            r = NotifyBuilder(title)
            self.notification_refs.append(r)
            return r
        return None



import logging


LOGGER = logging.getLogger(__name__)


class BaseDocViewController(object):
    def __init__(self, plugin):
        self.plugin = plugin

    def __str__(self):
        return str(type(self))

    def enter(self):
        LOGGER.debug("%s", self.enter)

    def exit(self):
        LOGGER.debug("%s", self.exit)

    def on_layout_size_allocate(self, layout):
        LOGGER.debug("%s(%s)", self.on_layout_size_allocate, layout)

    def on_page_size_obtained(self, page):
        LOGGER.debug("%s(%d)", self.on_page_size_obtained, page.page_idx)

    def on_vscroll(self, vadj):
        LOGGER.debug(
            "%s(%d, %d)",
            self.on_vscroll,
            vadj.get_value(),
            vadj.get_upper()
        )

    def on_vscroll_changed(self, vadj):
        self.on_vscroll(vadj)

    def on_vscroll_value_changed(self, vadj):
        self.on_vscroll(vadj)

    def doc_goto_page(self, page_idx):
        LOGGER.debug("%s(%d)", self.doc_goto_page, page_idx)

    def docview_set_layout(self, name):
        LOGGER.debug("%s(%s)", self.docview_set_layout, name)

    def docview_set_zoom(self, zoom):
        LOGGER.debug("%s(%f)", self.docview_set_zoom, zoom)

    def doc_reload(self):
        LOGGER.debug("%s()", self.doc_reload)

    def on_page_activated(self, page):
        LOGGER.debug("%s(%d)", self.on_page_activated, page.page_idx)

    def on_drag_motion(self, drag_context, x, y, time):
        LOGGER.debug("%s()", self.on_drag_motion)

    def on_drag_leave(self, drag_context, time):
        LOGGER.debug("%s()", self.on_drag_leave)

    def drag_and_drop_get_destination(self, x, y):
        LOGGER.debug("%s(%d, %d)", self.drag_and_drop_get_destination, x, y)

    def on_draw(self, cairo_context):
        # no LOGGER.debug() here for performance reasons
        pass

    def on_overlay_draw(self, overlay_drawing_area, cairo_context):
        # no LOGGER.debug() here for performance reasons
        pass

    def on_close(self):
        LOGGER.debug("%s()", self.on_close)



import logging

import PIL
import PIL.Image
import pillowfight

import openpaperwork_core

from . import AbstractImgEditor


LOGGER = logging.getLogger(__name__)


class CropImgEditor(AbstractImgEditor):
    def __init__(self, frame=None, **kwargs):
        self.frame = frame

    def transform(self, img, preview=False):
        if self.frame is None:
            # optimize the default frame
            LOGGER.info("Guessing new default cropping frame ...")

            # It's actually faster to resize down the image than look
            # for the scan borders on the full-size image.
            smaller_img = img.resize(
                (int(img.size[0] / 2), int(img.size[1] / 2)),
                getattr(PIL.Image, 'Resampling', PIL.Image).LANCZOS
            )
            frame = pillowfight.find_scan_borders(smaller_img)

            if frame[0] >= frame[2] or frame[1] >= frame[3]:
                LOGGER.info("Failed to guess a cropping frame")
                # defaulting to the whole image
                self.frame = (0, 0, img.size[0], img.size[1])
            else:
                # if we have found a valid frame
                LOGGER.info("Guessed cropping frame: %s", frame)
                self.frame = (
                    frame[0] * 2, frame[1] * 2, frame[2] * 2, frame[3] * 2
                )

        if preview:
            # we do not crop the preview. The UI displays a frame on top
            # of it showing where the cropping will happen.
            return img

        img = img.crop(self.frame)
        return img


class Plugin(openpaperwork_core.PluginBase):
    NAME = 'cropping'

    def get_interfaces(self):
        return ['img_editor']

    def img_editor_get_names(self, out: list):
        out.append(self.NAME)

    def img_editor_get(self, name, *args, **kwargs):
        if name != self.NAME:
            return None
        return CropImgEditor(**kwargs)

    def img_editor_set(self, inout: list, name, *args, **kwargs):
        if name != self.NAME:
            return None
        c = CropImgEditor(**kwargs)
        try:
            # Check if we already have a CropImgEditor in the list.
            # If so, do nothing
            inout.index(c)
        except ValueError:
            inout.append(c)

    def img_editor_unset(self, inout: list, name):
        if name != self.NAME:
            return None
        c = CropImgEditor()
        while c in inout:
            inout.remove(c)



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import sys

import openpaperwork_core
import openpaperwork_core.promise

import paperwork_backend.docexport

from . import util
from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.console = None

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "document_storage",
                "defaults": ["paperwork_backend.model.workdir"],
            },
            {
                "interface": "fs",
                "defaults": ["openpaperwork_gtk.fs.gio"],
            },
            {
                "interface": "export_pipes",
                "defaults": [
                    'paperwork_backend.docexport.generic',
                    'paperwork_backend.docexport.img',
                    'paperwork_backend.docexport.pdf',
                    'paperwork_backend.docexport.pillowfight',
                ],
            },
        ]

    def cmd_set_console(self, console):
        self.console = console

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'export', help=_(
                "Export a document, a page, or a set of pages."
                " Example:"
                " paperwork-cli export 20150303_2314_39 -p 2 -f img_boxes"
                " -f grayscale -f jpeg -o ~/tmp/pouet.jpg"
            )
        )
        p.add_argument('doc_id', help=_('Document to export'))
        p.add_argument(
            '--pages', '-p', type=str, required=False,
            help=_(
                "Pages to export"
                " (single integer, range or comma-separated list,"
                " default: all pages)"
            )
        )
        p.add_argument(
            '--filters', '-f', nargs=1, action='append',
            type=str, required=False,
            help=_(
                "Export filters. Specify this option once for each filter"
                " to apply (ex: '-f grayscale -f jpeg')."
            )
        )
        p.add_argument(
            '--out', '-o', type=str, required=False,
            help=_(
                "Output file/directory. If not specified, will list"
                " the filters that could be chained after those already"
                " specified."
            )
        )

    def _print_possible_pipes(self, next_pipes):
        next_pipes = [pipe.name for pipe in next_pipes]
        next_pipes.sort()
        self.console.print(_("Next possible filters are:"))
        for pipe in next_pipes:
            self.console.print(f"- {pipe}")
        return next_pipes

    @staticmethod
    def _data_type_to_str(dtype):
        if dtype == paperwork_backend.docexport.ExportDataType.DOCUMENT_SET:
            return _("a document list")
        elif dtype == paperwork_backend.docexport.ExportDataType.DOCUMENT:
            return _("a document")
        elif dtype == paperwork_backend.docexport.ExportDataType.PAGE:
            return _("pages")
        elif dtype == paperwork_backend.docexport.ExportDataType.IMG_BOXES:
            return _("Images and text boxes")
        return f"Unknown: {dtype}"

    def _check_pipeline(self, doc_url, pages, filters):
        filters = list(filters)

        next_pipes = []
        if pages is not None and len(pages) > 0:
            self.core.call_all(
                "export_get_pipes_by_page", next_pipes, doc_url, pages[0]
            )
        else:
            self.core.call_all(
                "export_get_pipes_by_doc_url", next_pipes, doc_url
            )
        if len(filters) <= 0:
            self.console.print(_("Need at least one filter."))
            return self._print_possible_pipes(next_pipes)
        if not filters[0] in next_pipes:
            sys.stderr.write(
                _("First filter cannot be '{}'").format(
                    filters[0].name
                ) + "\n"
            )
            self._print_possible_pipes(next_pipes)
            return False

        output_type = None
        while len(filters) > 0:
            f = filters.pop(0)
            if output_type is not None and f.input_type != output_type:
                sys.stderr.write(
                    _(
                        "Filter mismatch:"
                        " {0} expects {1} as input. Got {2} instead"
                    ).format(
                        f.name,
                        self._data_type_to_str(f.input_type),
                        self._data_type_to_str(output_type),
                    ) + "\n"
                )
                return False
            output_type = f.output_type

        expected_output_type = (
            paperwork_backend.docexport.ExportDataType.OUTPUT_URL_FILE
        )
        if output_type != expected_output_type:
            self.console.print(
                _("Last filter will output {}.").format(
                    self._data_type_to_str(output_type)
                ) + "\n"
            )
            next_pipes = []
            self.core.call_all(
                "export_get_pipes_by_input", next_pipes, output_type
            )
            return self._print_possible_pipes(next_pipes)
        return None

    def cmd_run(self, console, args):
        if args.command != 'export':
            return None

        # argument parsing

        doc_id = args.doc_id
        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        if doc_url is None:
            sys.stderr.write(f"Document {doc_id} doesn't exist\n")
            return False
        pages = util.parse_page_list(args)
        filters = args.filters if args.filters is not None else []
        out = args.out

        if pages is None or len(pages) <= 0:
            input_value = paperwork_backend.docexport.ExportData.build_doc(
                doc_id, doc_url
            )
        else:
            input_value = paperwork_backend.docexport.ExportData.build_pages(
                doc_id, doc_url, pages
            )

        filters_str = [f[0] for f in filters]
        filters = []
        for f in filters_str:
            p = self.core.call_success('export_get_pipe_by_name', f)
            if p is None:
                sys.stderr.write(
                    (_("Unknown filters: %s") % f) + "\n"
                )
                if len(filters) <= 0:
                    sys.stderr.write(
                        _(
                            "Run the command without any filter to have a list"
                            " of possible filters to start with."
                        ) + "\n"
                    )
                return False
            filters.append(p)

        r = self._check_pipeline(doc_url, pages, filters)
        if r is not None:
            return r

        if out is None or out == "":
            sys.stderr.write(
                _(
                    "Filter list is complete,"
                    " but no output file specified (-o)"
                ) + "\n"
            )
            return []

        # else try to export
        out = self.core.call_success("fs_safe", out)

        def get_input_value():
            return input_value

        promise = openpaperwork_core.promise.Promise(
            self.core, get_input_value
        )
        for pipe in filters:
            promise = promise.then(
                pipe.get_promise(result='final', target_file_url=out)
            )

        self.console.print(_("Exporting to %s ... ") % out)
        self.core.call_one("mainloop_schedule", promise.schedule)
        self.core.call_all("mainloop_quit_graceful")
        self.core.call_one("mainloop")
        self.console.print(_("Done") + "\n")
        r = (self.core.call_success("fs_exists", out) is not None)
        if not r:
            sys.stderr.write(_("Export failed !") + "\n")
        return r



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # noqa E402


DOC_BOX_XML = """
<?xml version="1.0" encoding="utf-8"?>
<interface domain="paperwork_gtk">
  <requires lib="gtk+" version="3.20"/>
  <template class="DocBox" parent="GtkListBoxRow">
    <property name="visible">True</property>
    <style>
      <class name="doclist_item"/>
    </style>
    <child>
      <object class="GtkBox" id="internal_doc_box">
        <property name="visible">True</property>
        <property name="can_focus">False</property>
        <property name="orientation">horizontal</property>
        <style>
          <class name="border-bottom-light"/>
        </style>
        <child>
          <object class="GtkOverlay">
            <property name="visible">True</property>
            <property name="can_focus">False</property>
            <child type="overlay">
              <object class="GtkCheckButton" id="doc_box_selector">
                <property name="visible">False</property>
                <property name="can_focus">False</property>
                <property name="relief">none</property>
                <property name="valign">end</property>
                <property name="margin_left">2</property>
                <property name="margin_right">2</property>
                <property name="margin_top">2</property>
                <property name="margin_bottom">2</property>
              </object>
              <packing>
                <property name="pass_through">True</property>
              </packing>
            </child>
            <child>
              <object class="GtkImage" id="doc_thumbnail">
              <property name="visible">False</property>
              <property name="icon_name">x-office-document-symbolic</property>
              <property name="icon_size">3</property>
              <style>
                <class name="doclist_thumbnail"/>
              </style>
              </object>
            </child>
          </object>
          <packing>
            <property name="pack_type">start</property>
          </packing>
        </child>

        <child>
          <object class="GtkViewport" id="layout_bin">
            <property name="visible">True</property>
            <property name="can_focus">False</property>
          </object>
          <packing>
            <property name="pack_type">start</property>
            <property name="expand">True</property>
            <property name="fill">True</property>
          </packing>
        </child>

        <child>

              <object class="GtkBox" id="doc_actions">
                <property name="visible">False</property>
                <property name="can_focus">False</property>
                <property name="orientation">vertical</property>

                <child>
                  <object class="GtkMenuButton" id="doc_actions_menu">
                    <property name="use_action_appearance">False</property>
                    <property name="visible">True</property>
                    <property name="can_focus">False</property>
                    <property name="use_underline">False</property>
                    <property name="relief">none</property>
                    <child>
                      <object class="GtkImage">
                      <property name="visible">True</property>
                      <property name="can_focus">False</property>
                      <property name="icon_name">view-more-symbolic</property>
                      <property name="icon_size">1</property>
                      </object>
                    </child>
                  </object>
                  <packing>
                    <property name="pack_type">end</property>
                    <property name="expand">True</property>
                    <property name="fill">True</property>
                  </packing>
                </child>

              </object>
          <packing>
            <property name="pack_type">start</property>
            <property name="expand">False</property>
            <property name="fill">True</property>
          </packing>
        </child>

      </object>
    </child>
  </template>
</interface>
"""


@Gtk.Template(string=DOC_BOX_XML)
class DocBox(Gtk.ListBoxRow):
    __gtype_name__ = "DocBox"

    box = Gtk.Template.Child("internal_doc_box")
    selector = Gtk.Template.Child("doc_box_selector")
    thumbnail = Gtk.Template.Child("doc_thumbnail")
    main_actions = Gtk.Template.Child("doc_actions")
    action_menu = Gtk.Template.Child("doc_actions_menu")
    layout_bin = Gtk.Template.Child("layout_bin")


DOC_MAIN_ACTION_XML = """
<?xml version="1.0" encoding="utf-8"?>
<interface domain="paperwork_gtk">
  <requires lib="gtk+" version="3.20"/>
  <template class="DocMainAction" parent="GtkButton">
    <property name="use_action_appearance">False</property>
    <property name="visible">True</property>
    <property name="can_focus">False</property>
    <property name="use_underline">False</property>
    <property name="relief">none</property>
    <property name="has_tooltip">True</property>
    <property name="tooltip_text"></property>
    <signal name="clicked" handler="on_clicked" swapped="no" />
    <child>
      <object class="GtkImage" id="doc_main_action_image">
        <property name="visible">True</property>
        <property name="can_focus">False</property>
      </object>
    </child>
  </template>
</interface>
"""


@Gtk.Template(string=DOC_MAIN_ACTION_XML)
class DocMainAction(Gtk.Button):
    __gtype_name__ = "DocMainAction"

    img = Gtk.Template.Child("doc_main_action_image")

    def __init__(self, text, icon_name, callback, *args, **kwargs):
        self.callback = callback
        super().__init__(*args, **kwargs)
        self.set_tooltip_text(text)
        self.img.set_from_icon_name(icon_name, Gtk.IconSize.MENU)

    @Gtk.Template.Callback("on_clicked")
    def on_clicked(self, button):
        self.callback()



import logging

try:
    from gi.repository import Gio
    GIO_AVAILABLE = True
except (ImportError, ValueError):
    GIO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = None
        self.active_page = None
        self.button_edit = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'doc_open',
            'page_actions',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview_pageinfo',
                'defaults': ['paperwork_gtk.mainwindow.docview.pageinfo'],
            },
            {
                'interface': 'gtk_page_editor',
                'defaults': ['paperwork_gtk.mainwindow.pageeditor'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageinfo",
            "actions.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.menu_model = self.widget_tree.get_object("page_menu_model")
        self.submenus = {}

        self.button_edit = self.widget_tree.get_object("page_action_edit")
        self.button_edit.connect(
            "clicked", self._on_edit
        )

        self.core.call_success(
            "page_info_add_right", self.widget_tree.get_object("page_actions")
        )

        self.core.call_success(
            "mainloop_schedule", self.core.call_all,
            "on_page_menu_ready"
        )

    def chkdeps(self, out: dict):
        if not GIO_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def on_page_shown(self, page_idx):
        self.active_page = page_idx

    def _on_edit(self, button):
        self.core.call_all(
            "gtk_open_page_editor", *self.active_doc, self.active_page
        )

    def page_menu_open(self):
        self.widget_tree.get_object("page_actions_other").clicked()

    def page_menu_append_item(self, item, submenu_name=None):
        menu = self.menu_model
        if submenu_name is not None:
            menu = self.submenus.get(submenu_name, None)
            if menu is None:
                menu = Gio.Menu()
                self.submenus[submenu_name] = menu
                self.menu_model.append_submenu(submenu_name, menu)
        menu.append_item(item)

    def screenshot_snap_all_doc_widgets(self, out_dir):
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("page_actions"),
            self.core.call_success("fs_join", out_dir, "page_actions.png"),
            margins=(50, 50, 50, 50)
        )

    def screenshot_snap_page_action_menu(self, out_file):
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("page_actions"),
            out_file,
            margins=(50, 200, 50, 50)
        )



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import shutil

import PIL
import PIL.Image
import rich.text

import openpaperwork_core

from .. import _


class AppendableImage(object):
    def __init__(self, width, expected_height):
        # When scanning, width is reliable.
        # However height is not reliable, and truncating an image is much
        # easier then increasing its size. --> we allocate a big image and will
        # truncate it when needed
        self.img = PIL.Image.new(
            "RGB", (width, 10 * expected_height)
        )
        self.width = width
        self.height = 0

    def append(self, chunk):
        assert self.img.size[0] == chunk.size[0]
        self.img.paste(chunk, (0, self.height))
        self.height += chunk.size[1]

    def get_image(self, start_line=0, end_line=None):
        if end_line is None:
            end_line = self.height
        return self.img.crop((0, start_line, self.img.size[0], end_line))


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.img = None
        self.last_line_displayed = 0  # in pixels
        self.txt_line_height = 0  # in pixels
        self.terminal_size = (80, 25)
        self.doc_id = None
        self.doc_url = None
        self.doc_renderer = None
        self.page_hashes = {}
        self.console = None

    def get_deps(self):
        return [
            # if there are no doc_renderer loaded, nothing it displayed, which
            # may be fine.
            #
            # (see paperwork-json)
            # {
            #     'interface': 'doc_renderer',
            #     'defaults': [],
            # },
            {
                'interface': 'img_renderer',
                'defaults': ['paperwork_shell.display.docrendering.img'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
        ]

    def cmd_set_console(self, console):
        self.console = console

    def on_scan_feed_start(self, scan_id):
        pass

    def on_scan_page_start(self, scan_id, page_nb, scan_params):
        self.console.print("")
        self.console.print(
            _("Scanning page {} (expected size: {}x{}) ...").format(
                page_nb + 1, scan_params.get_width(), scan_params.get_height()
            )
        )
        self.img = AppendableImage(
            scan_params.get_width(), scan_params.get_height()
        )
        self.last_line_displayed = 0

        # We need to figure out how many pixels one line of character will
        # represent.
        self.terminal_size = (
            # leave one character at the end of each line to make sure the
            # terminal doesn't add line return
            shutil.get_terminal_size()[0] - 1,
            shutil.get_terminal_size()[1]
        )
        ratio = scan_params.get_width() / self.terminal_size[0]
        # When Fabulous resizes an image to turn it into terminal characters,
        # it assumes that each characters is 1 pixel wide and 2 pixels high.
        self.txt_line_height = int(2 * ratio) + 1

    def on_scan_chunk(self, scan_id, scan_params, img_chunk):
        self.img.append(img_chunk)
        current_usable_line = (
            self.img.height - (self.img.height % self.txt_line_height)
        )
        if current_usable_line > self.last_line_displayed:
            img = self.img.get_image(
                self.last_line_displayed, current_usable_line
            )
            img = self.core.call_success(
                "img_render", img, terminal_width=self.terminal_size[0]
            )
            for line in img:
                self.console.print(rich.text.Text.from_ansi(line))
            self.last_line_displayed = current_usable_line

    def on_scan_page_end(self, scan_id, page_nb, img):
        img_size = img.size
        self.console.print(
            _("Page {} scanned (actual size: {}x{})").format(
                page_nb + 1, img_size[0], img_size[1]
            )
        )

    def on_scan_feed_end(self, scan_id):
        self.console.print("")
        self.console.print(_("End of paper feed"))

    def on_scan2doc_start(self, scan_id, doc_id, doc_url):
        self.doc_id = doc_id
        self.doc_url = doc_url
        renderers = []
        self.core.call_all("doc_renderer_get", renderers)
        self.doc_renderer = renderers[-1]
        self.page_hashes = {}

    def _compute_page_hash(self, doc_url, page_idx):
        return self.core.call_success(
            "page_get_hash_by_url", doc_url, page_idx
        )

    def on_scan2doc_page_scanned(self, scan_id, doc_id, doc_url, page_idx):
        self.console.print(_("Page {} in document {} created").format(
            page_idx, doc_id
        ))
        self.page_hashes[page_idx] = self._compute_page_hash(
            self.doc_url, page_idx
        )

    def on_scan2doc_end(self, scan_id, doc_id, doc_url):
        self._show_last_page()
        self.doc_id = None
        self.doc_url = None
        self.doc_renderer = None

    def _show_last_page(self):
        if self.doc_renderer is None:
            return

        for page_idx in self.page_hashes:
            # we only want to display the page if something has actually
            # changed
            page_hash = self._compute_page_hash(self.doc_url, page_idx)
            if self.page_hashes[page_idx] == page_hash:
                continue
            self.page_hashes[page_idx] = page_hash

            lines = self.doc_renderer.get_preview_output(
                self.doc_id, self.doc_url, self.terminal_size, page_idx
            )
            self.console.print("")
            for line in lines:
                self.console.print(rich.text.Text.from_ansi(line))
            self.console.print("")

    def on_progress(self, upd_type, progress, description=None):
        self._show_last_page()



import logging


LOGGER = logging.getLogger(__name__)


def _shift_pages(core, page_filename_fmt, doc_url, start_page_idx, offset):
    assert offset != 0

    total_pages = core.call_success("doc_get_nb_pages_by_url", doc_url)
    if total_pages is None:
        total_pages = 0

    if offset < 0:
        rng = range(start_page_idx + 1, total_pages + 1)
    elif offset > 0:
        rng = range(total_pages - 1, start_page_idx - 1, -1)

    for page_idx in rng:
        old_url = core.call_success(
            "fs_join", doc_url, page_filename_fmt.format(page_idx + 1)
        )
        if core.call_success("fs_exists", old_url) is None:
            continue
        new_url = core.call_success(
            "fs_join", doc_url, page_filename_fmt.format(
                page_idx + 1 + offset
            )
        )
        LOGGER.info("  - %s --> %s", old_url, new_url)
        core.call_success("fs_rename", old_url, new_url)


def delete_page_file(core, page_filename_fmt, doc_url, page_idx, trash=True):
    file_url = core.call_success(
        "fs_join", doc_url, page_filename_fmt.format(page_idx + 1)
    )
    if core.call_success("fs_exists", file_url) is not None:
        LOGGER.info(
            "(%s) Deleting %s p%d:", page_filename_fmt, doc_url, page_idx
        )
        core.call_success("fs_unlink", file_url, trash=trash)

    # move all the other pages 1 level down
    _shift_pages(core, page_filename_fmt, doc_url, page_idx, -1)
    return True


def move_page_file(
            core, page_filename_fmt,
            source_doc_url, source_page_idx,
            dest_doc_url, dest_page_idx
        ):
    if core.call_success("fs_exists", dest_doc_url) is None:
        assert dest_page_idx == 0
        core.call_success("fs_mkdir_p", dest_doc_url)

    LOGGER.info(
        "(%s) Move %s p%d --> %s p%d:", page_filename_fmt,
        source_doc_url, source_page_idx,
        dest_doc_url, dest_page_idx
    )

    # source_doc_url and dest_doc_url can be the same document, making
    # this change a little bit tricky. The simplest way to handle all the cases
    # is to:
    # --> move the page out of the source document (we rename it temporarily
    # page_filename_fmt.format(0))
    # --> then to insert it in the destination document

    # Move the page out of the source document
    src = core.call_success(
        "fs_join", source_doc_url,
        page_filename_fmt.format(source_page_idx + 1)
    )
    dst = core.call_success(
        "fs_join", dest_doc_url, page_filename_fmt.format(0)
    )
    if core.call_success("fs_exists", src) is not None:
        LOGGER.info("  - %s --> %s", src, dst)
        core.call_success("fs_rename", src, dst)

    _shift_pages(core, page_filename_fmt, source_doc_url, source_page_idx, -1)

    # Move the page in the destination document
    src = dst
    dst = core.call_success(
        "fs_join", dest_doc_url, page_filename_fmt.format(dest_page_idx + 1)
    )
    _shift_pages(core, page_filename_fmt, dest_doc_url, dest_page_idx, 1)

    if core.call_success("fs_exists", src) is not None:
        LOGGER.info("  - %s --> %s", src, dst)
        core.call_success("fs_rename", src, dst)
    return True


def get_nb_pages(core, doc_url, filename_regex):
    if core.call_success("fs_exists", doc_url) is None:
        return None
    if core.call_success("fs_isdir", doc_url) is None:
        return None
    files = core.call_success("fs_listdir", doc_url)
    if files is None:
        return None
    nb_pages = -1
    for f in files:
        f = core.call_success("fs_basename", f)
        match = filename_regex.match(f)
        if match is None:
            continue
        nb_pages = max(nb_pages, int(match.group(1)))
    if nb_pages <= 0:
        return None
    return nb_pages


def get_page_hash(core, doc_url, page_idx, filename_fmt):
    page_url = core.call_success(
        "fs_join", doc_url, filename_fmt.format(page_idx + 1)
    )
    if core.call_success("fs_exists", page_url) is None:
        return None
    return core.call_success("fs_hash", page_url)


def get_page_mtime(core, doc_url, page_idx, filename_fmt):
    page_url = core.call_success(
        "fs_join", doc_url, filename_fmt.format(page_idx + 1)
    )
    if core.call_success("fs_exists", page_url) is None:
        return
    return core.call_success("fs_get_mtime", page_url)


def get_doc_mtime(core, doc_url, filename_regex):
    r = -1
    if core.call_success("fs_exists", doc_url) is None:
        return
    if core.call_success("fs_isdir", doc_url) is None:
        return
    files = core.call_success("fs_listdir", doc_url)
    if files is None:
        return None
    for f in files:
        f = core.call_success("fs_basename", f)
        match = filename_regex.match(f)
        if match is None:
            continue
        page_url = core.call_success("fs_join", doc_url, f)
        r = max(r, core.call_success("fs_get_mtime", page_url))
    if r < 0:
        return None
    return r



import logging

import openpaperwork_core
import openpaperwork_core.promise


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -2000000

    def __init__(self):
        super().__init__()
        self.widget_tree = None

    def get_interfaces(self):
        return [
            'doc_open',
            'gtk_search_field_completion',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_search_field',
                'defaults': ['paperwork_gtk.mainwindow.search.field'],
            },
            {
                'interface': 'suggestions',
                'defaults': ['paperwork_backend.index.whoosh'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.search", "suggestions.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        suggestion_revealer = self.widget_tree.get_object(
            "suggestions_revealer"
        )
        self.core.call_all("search_field_add", suggestion_revealer)

        self.widget_tree.get_object("button_close").connect(
            "clicked", self._close
        )
        self.widget_tree.get_object("treeview_suggestions").connect(
            "row-activated", self._on_row_activated
        )

    def _get_suggestions(self, query):
        out = set()
        self.core.call_all("suggestion_get", out, query)
        out = list(out)
        out.sort()
        return out

    def _show_suggestions(self, suggestions, query):
        if query != self.core.call_success("search_get"):
            # Text has changed while we were looking for suggestions
            # No point in displaying them now.
            return
        LOGGER.info("%d suggestions found", len(suggestions))
        model = self.widget_tree.get_object("liststore_suggestions")
        model.clear()
        for suggestion in suggestions:
            model.append((suggestion,))
        self.widget_tree.get_object(
            "suggestions_revealer"
        ).set_reveal_child(len(suggestions) > 0)

    def search_by_keywords(self, query):
        self.widget_tree.get_object("liststore_suggestions").clear()

        self.widget_tree.get_object(
            "suggestions_revealer"
        ).set_reveal_child(False)

        if query == "":
            return

        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self._get_suggestions, args=(query,)
        )
        promise = promise.then(self._show_suggestions, query)
        self.core.call_success(
            "work_queue_add_promise", "doc_search", promise
        )

    def _close(self, *args, **kwargs):
        self.widget_tree.get_object(
            "suggestions_revealer"
        ).set_reveal_child(False)

    def doc_open(self, *args, **kwargs):
        self._close()

    def _on_row_activated(self, treeview, path, column):
        model = treeview.get_model()
        suggestion = model[path][0]
        self.core.call_all("search_set", suggestion)



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.active_doc = None

    def get_interfaces(self):
        return [
            'gtk_doc_property',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_calendar_popover',
                'defaults': ['openpaperwork_gtk.widgets.calendar'],
            },
            {
                'interface': 'gtk_doc_properties',
                'defaults': ['paperwork_gtk.docproperties'],
            },
            {
                'interface': 'gtk_colors',
                'defaults': ['openpaperwork_gtk.colors'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def doc_properties_components_get(self, out: list, multiple_docs=False):
        if multiple_docs:
            return
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docproperties", "name.glade"
        )

        self.core.call_all(
            "gtk_calendar_add_popover",
            self.widget_tree.get_object("docname_entry")
        )

        self.widget_tree.get_object("docname_entry").connect(
            "changed", self._on_doc_date_changed
        )

        out.append(self.widget_tree.get_object("docname"))

    def doc_properties_components_set_active_doc(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

        doc_date = self.core.call_success("doc_get_date_by_id", doc_id)
        if doc_date is not None:
            doc_txt = self.core.call_success("i18n_date_short", doc_date)
        else:
            doc_txt = doc_id
        self.widget_tree.get_object("docname_entry").set_text(doc_txt)

    def _check_doc_id(self, doc_id):
        if doc_id == "":
            return False
        for forbidden in ["/", "\\", "?", "*"]:
            if forbidden in doc_id:
                return False
        return True

    def _on_doc_date_changed(self, gtk_entry):
        txt = gtk_entry.get_text()
        txt = txt.strip()
        r = self.core.call_success("i18n_parse_date_short", txt)
        if r is not None:
            self.core.call_all("gtk_entry_reset_colors", gtk_entry)
        elif not self._check_doc_id(txt):
            self.core.call_all("gtk_entry_set_colors", gtk_entry, bg="#ff0000")
        else:
            self.core.call_all("gtk_entry_set_colors", gtk_entry, bg="#ee9000")

    def doc_properties_components_apply_changes(self, out):
        if out.multiple_docs:
            return

        doc_id = self.widget_tree.get_object("docname_entry").get_text()
        doc_id = doc_id.strip()

        doc_date = self.core.call_success("i18n_parse_date_short", doc_id)
        if doc_date is not None:
            doc_id = self.core.call_success("doc_get_id_by_date", doc_date)
        elif not self._check_doc_id(doc_id):
            LOGGER.error(
                "Invalid document id specified. Won't rename the document"
            )
            return
        else:
            LOGGER.warning(
                "Failed to parse document date: %s. Using as is", doc_id
            )

        orig_id = out.doc_id
        orig_date = self.core.call_success("doc_get_date_by_id", orig_id)
        if orig_date is not None:
            orig_date = orig_date.date()
        else:
            orig_date = orig_id

        dest_id = doc_id
        dest_date = self.core.call_success("doc_get_date_by_id", dest_id)
        if dest_date is not None:
            dest_date = dest_date.date()
        else:
            dest_date = dest_id

        if orig_date == dest_date:
            return

        LOGGER.info("Previous document id: %s (%s)", orig_id, orig_date)
        LOGGER.info("New document id: %s (%s)", dest_id, dest_date)

        orig_url = self.core.call_success("doc_id_to_url", orig_id)
        dest_url = self.core.call_success(
            "doc_id_to_url", dest_id, existing=False
        )

        LOGGER.info("Renaming document %s into %s", orig_id, dest_id)
        dest_url = self.core.call_success(
            "doc_rename_by_url", orig_url, dest_url
        )
        if not dest_url:
            LOGGER.warning("Renaming %s into %s failed", orig_id, dest_id)
            return
        dest_id = self.core.call_success("doc_url_to_id", dest_url)

        out.del_docs.add(out.doc_id)
        out.new_docs.add(dest_id)
        out.doc_id = dest_id

    def doc_properties_components_cancel_changes(self):
        self.doc_properties_components_set_active_doc(*self.active_doc)



import logging
import random

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise
import openpaperwork_gtk.deps

import paperwork_backend.docexport

from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.ui = None
        self.active_doc = None
        self.active_page_idx = None
        self.windows = []

        self.button_validate = None
        self.button_send_email = None
        self.preciew = None
        self.zoom = None
        self.quality = None
        self.combobox_page_format = None
        self.model_page_format = None

        # The reference ('ref_*') is what is displayed as example to the user.
        # It's always only one page. The exported document size is estimated
        # by multiplying this reference page after post-processing by the
        # number of pages in the document.

        # Inputs (export_input_*) refer to export pipe inputs
        # (see paperwork_backend.docexport)

        self.ref_input_page = None
        self.ref_input_doc = None
        self.export_input_type = None
        self.export_input = None
        self.export_input_doc_urls = None
        self.need_zoom_auto = True

        self.pipeline = []
        self.renderer = None
        self.tmp_file_url = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'doc_open',
            'gtk_exporter',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'cairo_url',
                'defaults': [
                    'paperwork_backend.cairo.pillow',
                    'paperwork_backend.cairo.poppler',
                ],
            },
            {
                'interface': 'export_pipes',
                'defaults': [
                    'paperwork_backend.docexport.img',
                    'paperwork_backend.docexport.pdf',
                    'paperwork_backend.docexport.pillowfight',
                ],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_zoomable',
                'defaults': ['paperwork_gtk.gesture.zoom'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'notifications',
                'defaults': [
                    'paperwork_gtk.notifications.dialog',
                    'paperwork_gtk.notifications.notify',
                ],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
            {
                'interface': 'external_apps',
                'defaults': [
                    'openpaperwork_core.external_apps.dbus',
                    'openpaperwork_core.external_apps.windows',
                    'openpaperwork_core.external_apps.xdg',
                ],
            },
        ]

    def init(self, core):
        super().init(core)

        opt = self.core.call_success(
            "config_build_simple", "GUI", "last_export_uri",
            lambda: None
        )
        self.core.call_all("config_register", "last_export_uri", opt)

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.exporter", "exporter.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.widget_tree.get_object("exporter_cancel").connect(
            "clicked", self._on_cancel
        )
        self.button_validate = self.widget_tree.get_object("exporter_validate")
        self.button_validate.connect(
            "clicked", self._on_apply
        )
        self.button_send_email = self.widget_tree.get_object(
            "exporter_send_email"
        )
        self.button_send_email.connect(
            "clicked", self._on_send_email
        )
        can_send_as_attachment = core.call_success(
            "external_app_can_send_as_attachment"
        )
        if not can_send_as_attachment:
            self.button_send_email.set_visible(False)

        self.zoom = self.widget_tree.get_object("exporter_zoom_adjustment")
        self.zoom.connect("value-changed", self._on_zoom_changed)

        self.quality = self.widget_tree.get_object(
            "exporter_quality_adjustment"
        )
        self.quality.connect("value-changed", self._on_quality_changed)

        self.preview = self.widget_tree.get_object("exporter_img")
        self.preview.connect("draw", self._on_draw)

        self.core.call_all(
            "on_zoomable_widget_new",
            self.widget_tree.get_object("exporter_scroll"),
            self.zoom
        )

        self.core.call_all(
            "mainwindow_add", "right", "exporter", prio=0,
            header=self.widget_tree.get_object("exporter_header"),
            body=self.widget_tree.get_object("exporter_body")
        )

        self.core.call_success("work_queue_create", "exporter")

        self.combobox_page_format = self.widget_tree.get_object(
            "exporter_page_format"
        )
        self.model_page_format = self.widget_tree.get_object(
            "exporter_page_format_model"
        )

        self._add_page_formats()

        self.combobox_page_format.connect(
            "changed", self._on_page_format_changed
        )

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def on_page_shown(self, page_idx):
        self.active_page_idx = page_idx

    def _add_page_formats(self):
        default_idx = -1
        for (idx, paper_size) in enumerate(
                Gtk.PaperSize.get_paper_sizes(True)):
            store_data = (
                paper_size.get_display_name(),
                paper_size.get_width(Gtk.Unit.POINTS),
                paper_size.get_height(Gtk.Unit.POINTS)
            )
            self.model_page_format.append(store_data)
            if paper_size.get_name() == paper_size.get_default():
                default_idx = idx
        if default_idx >= 0:
            self.combobox_page_format.set_active(default_idx)

    def _get_possible_pipes(self, input_type, active_pipe=""):
        pipes = []
        if input_type == paperwork_backend.docexport.ExportDataType.DOCUMENT:
            self.core.call_all(
                "export_get_pipes_by_doc_urls",
                pipes, self.export_input_doc_urls
            )
        else:
            self.core.call_all("export_get_pipes_by_input", pipes, input_type)
        pipeline = {p.name for p in self.pipeline}
        if active_pipe in pipeline:
            pipeline.remove(active_pipe)
        pipes = [p for p in pipes if p.name not in pipeline]
        return pipes

    def _expand_pipeline(self):
        """
        When there is only once choice possible, this isn't really a choice.
        --> we expand automatically the pipeline to include the only choice
        possible.
        """
        while True:
            if len(self.pipeline) > 0:
                last_output_type = self.pipeline[-1].output_type
            else:
                last_output_type = self.export_input_type

            if last_output_type == 'file_url':
                # pipeline is complete
                return

            pipes = self._get_possible_pipes(last_output_type)
            if len(pipes) != 1:
                # there is a choice (or none at all), nothing to do
                return
            pipe = pipes[0]
            assert pipe.name not in (p.name for p in self.pipeline)
            self.pipeline.append(pipe)
            LOGGER.info("Pipeline expanded: %s", [
                p.name for p in self.pipeline
            ])

    def _add_combobox(self, steps, previous_input_type, active_pipe):
        possible_alternative_pipes = self._get_possible_pipes(
            previous_input_type, active_pipe
        )
        if len(possible_alternative_pipes) <= 1:
            # not really a choice
            return

        possible_alternative_pipes.sort(key=lambda p: str(p))

        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.exporter", "pipe.glade"
        )
        combobox = widget_tree.get_object("exporter_pipe")
        choices = widget_tree.get_object("exporter_pipe_model")
        choices.clear()
        choices.append((_("Please select the next export step"), ""))
        active_idx = 0
        for (alternative_idx, alternative_pipe) in enumerate(
                possible_alternative_pipes):
            choices.append(
                (str(alternative_pipe), alternative_pipe.name)
            )
            if alternative_pipe.name == active_pipe:
                active_idx = alternative_idx + 1
        combobox.set_active(active_idx)
        combobox.connect('changed', self._on_pipeline_changed)
        steps.add(combobox)

    def _refresh_pipeline_ui(self):
        LOGGER.info("Displaying pipeline: %s", [p.name for p in self.pipeline])
        steps = self.widget_tree.get_object("exporter_steps")
        for widget in steps:
            steps.remove(widget)

        for (pipe_idx, pipe) in enumerate(self.pipeline):
            if pipe_idx <= 0:
                previous_input_type = self.export_input_type
            else:
                previous_input_type = self.pipeline[pipe_idx - 1].output_type
            self._add_combobox(steps, previous_input_type, pipe.name)

        if len(self.pipeline) > 0:
            last_output_type = self.pipeline[-1].output_type
            LOGGER.info(
                "Last pipe: %s ; Last output type: %s",
                self.pipeline[-1], last_output_type
            )
        else:
            last_output_type = self.export_input_type
            LOGGER.info("Input type: %s", last_output_type)

        t = paperwork_backend.docexport.ExportDataType.OUTPUT_URL_FILE
        if last_output_type == t:
            self.button_validate.set_sensitive(True)
            self.button_send_email.set_sensitive(True)
        else:
            self._add_combobox(steps, last_output_type, "")
            self.button_validate.set_sensitive(False)
            self.button_send_email.set_sensitive(False)

    def _rebuild_pipeline_from_ui(self):
        LOGGER.info("Rebuilding pipeline from UI")
        self.pipeline = []
        steps = self.widget_tree.get_object("exporter_steps")
        for combobox in steps:
            self._expand_pipeline()
            pipe = combobox.get_active()
            pipe = combobox.get_model()[pipe][1]
            if pipe == "":
                return
            pipe = self.core.call_success("export_get_pipe_by_name", pipe)
            if len(self.pipeline) > 0:
                if self.pipeline[-1].output_type != pipe.input_type:
                    LOGGER.warning(
                        "Pipe [%s] cannot be followed [%s]. Ignoring"
                        " remainder of the pipeline",
                        self.pipeline[-1], pipe
                    )
                    break
            self.pipeline.append(pipe)
        LOGGER.info("Pipeline from UI: %s", [p.name for p in self.pipeline])

    def _hide_preview(self):
        if self.tmp_file_url is not None:
            self.core.call_success("fs_unlink", self.tmp_file_url, trash=False)
            self.tmp_file_url = None
        if self.renderer is not None:
            self.renderer.close()
            self.renderer = None
        label = self.widget_tree.get_object("exporter_estimated_size")
        label.set_text("")
        label.set_visible(False)

    def _resize_preview(self):
        self.preview.set_size_request(
            int(self.renderer.size[0] * self.renderer.zoom),
            int(self.renderer.size[1] * self.renderer.zoom)
        )

    def _on_size_obtained(self, *args, **kwargs):
        if self.need_zoom_auto:
            allocation = self.widget_tree.get_object(
                "exporter_scroll"
            ).get_allocation()
            allocation = (allocation.width - 20, allocation.height - 20)
            zoom = min(
                allocation[0] / self.renderer.size[0],
                allocation[1] / self.renderer.size[1],
            )
            self.zoom.set_value(zoom)
            self.renderer.zoom = zoom
            self.need_zoom_auto = False
        self._resize_preview()

    def _redraw_preview(self, renderer):
        self.preview.queue_draw()

    def _show_preview(self, tmp_file_urls):
        tmp_file_url = list(tmp_file_urls)[0]
        LOGGER.info("Preview: %s", tmp_file_url)
        self.tmp_file_url = tmp_file_url
        self.renderer = self.core.call_success(
            "cairo_renderer_by_url", "exporter", tmp_file_url
        )
        self.renderer.zoom = self.zoom.get_value()
        self.renderer.connect("size_obtained", self._on_size_obtained)
        self.renderer.connect("img_obtained", self._redraw_preview)
        self.renderer.start()
        self.renderer.render()
        return tmp_file_url

    def _show_estimated_size(self, tmp_file_url):
        preview_size = self.core.call_success("fs_getsize", tmp_file_url)
        factors = (
            p.get_estimated_size_factor(self.export_input)
            for p in self.pipeline
        )
        final_size = preview_size
        for f in factors:
            final_size *= f
        LOGGER.info(
            "Preview size: %d ; Estimated final size: %d",
            preview_size, final_size
        )
        final_size = self.core.call_success("i18n_file_size", final_size)
        label_txt = _("Estimated file size: %s") % (final_size)
        label = self.widget_tree.get_object("exporter_estimated_size")
        label.set_text(label_txt)
        label.set_visible(True)
        return tmp_file_url

    def _get_pipe_plug(self):
        if self.pipeline[-1].output_type == 'pages':
            return self.core.call_success(
                "export_get_pipe_by_name", "png"
            )
        return None

    def _set_quality(self):
        can_change_quality = False
        for pipe in self.pipeline:
            if not pipe.can_change_quality:
                continue
            can_change_quality = True
            pipe.set_quality(self.quality.get_value() / 100)

        self.widget_tree.get_object("exporter_quality").set_sensitive(
            can_change_quality
        )
        self.widget_tree.get_object("exporter_quality_label").set_sensitive(
            can_change_quality
        )

    def _set_page_format(self):
        page_format = self.combobox_page_format.get_active()
        page_format = self.model_page_format[page_format]

        can_change_page_format = False
        for pipe in self.pipeline:
            if not pipe.can_change_page_format:
                continue
            can_change_page_format = True
            pipe.set_page_format((page_format[1], page_format[2]))

        self.widget_tree.get_object("exporter_page_format").set_sensitive(
            can_change_page_format
        )
        self.widget_tree.get_object(
            "exporter_page_format_label"
        ).set_sensitive(
            can_change_page_format
        )

    def _reload_preview(self):
        self.core.call_all("work_queue_cancel_all", "exporter")

        promise = openpaperwork_core.promise.Promise(
            self.core, self._hide_preview
        )

        if len(self.pipeline) <= 0:
            self.core.call_success(
                "work_queue_add_promise", "exporter", promise
            )
            return

        # some pipeline only accepts ExportDataType.DOCUMENT as input,
        # other accept ExportDataType.PAGE, but not immediately.
        # For the preview, we prefer ExportDataType.PAGE,
        # if not available we fall back to ExportDataType.DOCUMENT
        for (idx, pipe) in enumerate(self.pipeline):
            if (pipe.input_type ==
                    paperwork_backend.docexport.ExportDataType.PAGE):
                pipeline = self.pipeline[idx:]
                ref_input = self.ref_input_page
                break
        else:
            for (idx, pipe) in enumerate(self.pipeline):
                if (pipe.input_type ==
                        paperwork_backend.docexport.ExportDataType.DOCUMENT):
                    pipeline = self.pipeline[idx:]
                    ref_input = self.ref_input_doc
                    break
            else:
                LOGGER.warning(
                    "Can't display export preview:"
                    " No matching input pipe found in %s",
                    [str(p) for p in self.pipeline]
                )
                return

        ref_input = ref_input.clone()

        pipe_plug = self._get_pipe_plug()
        if pipe_plug is not None:
            if pipe_plug.can_change_quality:
                pipe_plug.set_quality(self.quality.get_value() / 100)
            if pipe_plug.can_change_page_format:
                page_format = self.combobox_page_format.get_active()
                page_format = self.model_page_format[page_format]
                pipe_plug.set_page_format((page_format[1], page_format[2]))

        t = paperwork_backend.docexport.ExportDataType.OUTPUT_URL_FILE
        if pipe_plug is None and pipeline[-1].output_type != t:
            LOGGER.warning(
                "Can't display export preview: unexpected pipe end: %s",
                pipeline[-1].output_type
            )
            self.core.call_success(
                "work_queue_add_promise", "exporter", promise
            )
            return

        promise = promise.then(self.core.call_all, "on_busy")
        promise = promise.then(lambda *args, **kwargs: ref_input)
        for pipe in pipeline:
            promise = promise.then(pipe.get_promise(result='preview'))
        if pipe_plug is not None:
            promise = promise.then(pipe_plug.get_promise(result='preview'))
        promise = promise.then(self._show_preview)
        if pipe_plug is None:
            promise = promise.then(self._show_estimated_size)
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.core.call_all, "on_idle")
        self.core.call_success("work_queue_add_promise", "exporter", promise)

    def _gtk_open_exporter(self):
        self.pipeline = []
        self.need_zoom_auto = True

        self._expand_pipeline()
        self._refresh_pipeline_ui()
        self._set_quality()
        self._set_page_format()
        self._reload_preview()
        self.core.call_all("mainwindow_show", "right", "exporter")

    def gtk_open_exporter(self, doc_id, doc_url, page_idx=None):
        ref_page_idx = 0
        if page_idx is not None:
            ref_page_idx = page_idx
        elif doc_url == self.active_doc[1]:
            ref_page_idx = self.active_page_idx

        self.ref_input_page = (
            paperwork_backend.docexport.ExportData.build_page(
                doc_id, doc_url, ref_page_idx
            )
        )
        self.ref_input_doc = (
            paperwork_backend.docexport.ExportData.build_doc(doc_id, doc_url)
        )

        self.export_input_type = (
            paperwork_backend.docexport.ExportDataType.DOCUMENT
            if page_idx is None else
            paperwork_backend.docexport.ExportDataType.PAGE
        )
        self.export_input = (
            paperwork_backend.docexport.ExportData.build_doc(doc_id, doc_url)
            if page_idx is None else
            paperwork_backend.docexport.ExportData.build_page(
                doc_id, doc_url, page_idx
            )
        )
        self.export_input_doc_urls = [doc_url]
        self._gtk_open_exporter()

    def gtk_open_exporter_multiple_docs(
            self, docs, ref_doc_id, ref_doc_url, ref_page_idx):
        self.ref_input_page = (
            paperwork_backend.docexport.ExportData.build_page(
                ref_doc_id, ref_doc_url, ref_page_idx
            )
        )
        self.ref_input_doc = (
            paperwork_backend.docexport.ExportData.build_doc(
                ref_doc_id, ref_doc_url
            )
        )

        self.export_input_type = (
            paperwork_backend.docexport.ExportDataType.DOCUMENT_SET
        )
        self.export_input = (
            paperwork_backend.docexport.ExportData.build_doc_set(docs)
        )
        self.export_input_doc_urls = [doc[1] for doc in docs]
        self._gtk_open_exporter()

    def _on_draw(self, drawing_area, cairo_context):
        if self.renderer is None:
            return
        self.renderer.draw(cairo_context)

    def _on_zoom_changed(self, adj):
        if self.renderer is None:
            return
        self.renderer.zoom = adj.get_value()
        self._resize_preview()
        self.preview.queue_draw()

    def _on_quality_changed(self, adj):
        self._set_quality()
        self._reload_preview()

    def _on_page_format_changed(self, combobox):
        self._set_page_format()
        self._reload_preview()

    def _on_pipeline_changed(self, *args, **kwargs):
        self._rebuild_pipeline_from_ui()
        self._expand_pipeline()
        # WORKAROUND(JFlesch): call this method using mainloop_schedule.
        # Otherwise, in Flatpak, with Gnome 40, it crashes.
        self.core.call_one("mainloop_schedule", self._refresh_pipeline_ui)
        self._set_quality()
        self._set_page_format()
        self._reload_preview()

    def _on_cancel(self, button):
        LOGGER.info("Export canceled")
        self.core.call_all("mainwindow_back", side="right")

    def _on_apply(self, button):
        LOGGER.info("Export settings defined. Opening file chooser dialog")

        dialog = Gtk.FileChooserNative.new(
            _("Select a file to which export the document"),
            self.windows[-1],
            Gtk.FileChooserAction.SAVE,
            None, None
        )
        dialog.set_modal(True)
        dialog.set_local_only(False)

        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("Any files"))
        file_filter.add_pattern("*.*")
        dialog.add_filter(file_filter)

        (mime, file_extensions) = self.pipeline[-1].get_output_mime()
        file_filter = Gtk.FileFilter()
        file_filter.add_mime_type(mime)
        file_filter.set_name(file_extensions[0])  # TODO(Jflesch): better name
        dialog.add_filter(file_filter)
        dialog.set_filter(file_filter)

        last_export_uri = self.core.call_success(
            "config_get", "last_export_uri"
        )
        if last_export_uri is not None:
            # this filename does not exist, but the parent directory
            # will be selected. A hack inspired by what firefox does.
            dialog.set_uri(last_export_uri + str(random.randint(0, 10000)))

        dialog.connect("response", self._on_dialog_response)
        dialog.run()
        dialog.hide()
        dialog.destroy()

    def _on_dialog_response(self, dialog, response_id):
        if (response_id != Gtk.ResponseType.ACCEPT and
                response_id != Gtk.ResponseType.OK and
                response_id != Gtk.ResponseType.YES and
                response_id != Gtk.ResponseType.APPLY):
            LOGGER.info("User canceled (response_id=%d)", response_id)
            return

        selected = dialog.get_uris()[0]
        self.core.call_all("mainwindow_back", side="right")

        # make sure the file extension is set
        (mime, file_extensions) = self.pipeline[-1].get_output_mime()
        for file_extension in file_extensions:
            if selected.lower().endswith(file_extension):
                break
        else:
            selected += "." + file_extensions[0]

        self.core.call_all("config_put", "last_export_uri", selected)

        promise = openpaperwork_core.promise.Promise(
            self.core, lambda: self.export_input
        )
        for pipe in self.pipeline:
            promise = promise.then(pipe.get_promise(
                result='final', target_file_url=selected
            ))
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(Gtk.RecentManager().add_item, selected)
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.catch(self._on_error)

        # do not use the work queue ; must never be canceled
        promise.schedule()

    def _on_send_email(self, button):
        (_, file_extensions) = self.pipeline[-1].get_output_mime()
        (target_file_url, fd) = self.core.call_success(
            "fs_mktemp", suffix="." + file_extensions[0], on_disk=True
        )
        fd.close()  # we only need the name
        self.core.call_all("mainwindow_back", side="right")

        promise = openpaperwork_core.promise.Promise(
            self.core, lambda: self.export_input
        )
        for pipe in self.pipeline:
            promise = promise.then(pipe.get_promise(
                result='final', target_file_url=target_file_url
            ))
        promise = promise.then(lambda *args, **kwargs: self.core.call_success(
            "external_app_send_as_attachment", target_file_url
        ))
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.catch(self._on_error)

        # do not use the work queue ; must never be canceled
        promise.schedule()

    def _on_error(self, exc):
        LOGGER.error("Export failed", exc_info=exc)
        notif = self.core.call_success(
            "get_notification_builder",
            _("Export has failed"),
        )
        notif.set_message(f"Export has failed: {type(exc)} {exc}")
        notif.set_icon("dialog-error")
        notif.show()
